/*
 * @(#)BaseCharacterIterator.java	1.7 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package java.awt.font;

import java.text.CharacterIterator;

/**
 * A utility class that serves as the base of iterators over String and 
 * StyledString.
 * Subclasses override charAt(pos) to return the appropriate character at 
 * that position.
 */
abstract class BaseCharacterIterator implements CharacterIterator {
    private int begin;
    private int end;
    private int pos;

    protected abstract char charAt(int pos);

    public BaseCharacterIterator(int begin, int end, int pos) {
        if (begin < 0 || begin > end) {
            throw new IllegalArgumentException("Invalid substring range");
        }

        if (pos < begin || pos > end) {
            // why not begin - 1???
            throw new IllegalArgumentException("Invalid position");
        }
        
        this.begin = begin;
        this.end = end;
        this.pos = pos;
    }
    
    public char first() {
        pos = begin;
        return charAt(pos);
    }

    public char last() {
        pos = end - 1;
        return charAt(pos);
    }

    public char setIndex(int p) {
        if (p < begin || p > end) {
            // why not begin - 1???
            throw new IllegalArgumentException("Invalid index");
        }
        pos = p;
        return current();
    }

    public char current() {
        if (pos >= begin && pos < end) {
            return charAt(pos);
        } else {
            return DONE;
        }
    }

    public char next() {
        if (++pos < end) {
            return charAt(pos);
        } else {
            pos = end;
            return DONE;
        }
    }
    
    // ???
    // unfortunately, pinning to begin means that the behavior differs
    // at the beginning and end of the range.  Is this what we want?
    //
    // begin();
    // char c1 = previous();
    // char c2 = current();
    // if (c1 != c2) whoops;
    //
    // You'd think current should always equal the most recent result from
    // previous, next, begin, or end.
    //
    // Wouldn't we also want the iterator to be well-behaved on both ends?
    // i.e.
    // char c1 = begin(); previous(); if (c1 != next()) whoops;
    // this is how it works at the end.
    //
    public char previous() {
        if (--pos >= begin) {
            return charAt(pos);
        } else {
            pos = begin; // not begin - 1?
            return DONE;
        }
    }

    public int getBeginIndex() {
        return begin;
    }
    
    public int getEndIndex() {
        return end;
    }

    public int getIndex() {
        return pos;
    }

    public boolean equals(Object obj) {
        return (obj instanceof BaseCharacterIterator) &&
            equals((BaseCharacterIterator)obj);
    }

    public boolean equals(BaseCharacterIterator iter) {
        return iter != null && pos == iter.pos && begin == iter.begin &&
            end == iter.end;
    }

    public int hashCode() {
        return pos ^ begin ^ end;
    }

    public Object clone() {
    	try {
            return super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
    }

    public void restrict(int start, int limit) {
        begin = Math.max(start, begin);
        end = Math.min(end, limit);
        if (begin >= end) {
            begin = end;
        }
        if (begin > pos) {
            pos = begin;
        } else if (end < pos) {
            pos = end;
        }
    }
}
