/*
 * @(#)AffineTransformOp.java	1.28 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;
import java.awt.Rectangle;

/**
 * This abstract class uses an affine transform to perform a
 * linear mapping from 2D coordinates in the source image or Raster to
 * 2D coordinates in the destination image or Raster.
 * The subclasses define what kind of interpolation will be done when
 * the image is transformed.  These subclasses include:
 * BilinearAffineTransformOp and NearestNeighborAffineTransformOp.
 * This class of objects can be passed into a BufferedImageFilter to
 * operate on a BufferedImage within the Image Producer/
 * Consumer model.
 * <p>
 * Note that the following constraints have to be met:
 * <li>Source and Destination must be different.
 * <li>For Rasters, the number of Channels in the Source must be equal
 * to the number of Channels in the Destination.
 * @see AffineTransform
 * @see BilinearAffineTransformOp
 * @see NearestNeighborAffineTransformOp
 * @see BufferedImageFilter
 * @version 10 Feb 1997
 */
public abstract class AffineTransformOp implements BufferedImageOp, RasterOp {
    private AffineTransform xform;

    /**
     * Constructs an AffineTransformOp given an affine transform.
     */
    protected AffineTransformOp(AffineTransform xform) {
        this.xform = xform;
    }

    /**
     * Abstract method to transform the BufferedImage.  
     * If the color models for the two images do not match, a color
     * conversion into the destination color model will be performed.
     * If the destination image is null,
     * a BufferedImage will be created with the source ColorModel.
     * The IllegalArgumentException may be thrown if the source is
     * the same as the destination.
     */
    public abstract BufferedImage filter(BufferedImage src,
                                         BufferedImage dest);

    /**
     * Abstract method to transform the Raster.  This operation will
     * perform the transform channel by channel.
     * If the destination Raster is null, a new Raster will be created.
     * The IllegalArgumentException may be thrown if the source is
     * the same as the destination or if the number of channels in
     * the source is not equal to the number of channels in the
     * destination.
     */
    public abstract WritableRaster filter(Raster src, WritableRaster dest);

    /**
     * Returns the bounding box of the transformed destination.
     */
    public Rectangle2D getDestBounds (BufferedImage src) {
        return getDestBounds(src.getRaster());
    }

    /**
     * Returns the bounding box of the transformed destination.  
     */
    public Rectangle2D getDestBounds (Raster src) {
        int w = src.getWidth();
        int h = src.getHeight();

        // Get the bounding box of the src and transform the corners
        float[] pts = {0, 0, w, 0, w, h, 0, h};
        xform.transform(pts, 0, pts, 0, 4);

        // Get the min, max of the dst
        float fmaxX = pts[0];
        float fmaxY = pts[1];
        int maxX;
        int maxY;
        for (int i=0; i < 8; i+=2) {
            if (pts[i] > fmaxX) {
                fmaxX = pts[i];
            }
            if (pts[i+1] > fmaxY) {
                fmaxY = pts[i+1];
            }
        }

        maxX = (fmaxX <= 0 ? 0 : (int) Math.ceil(fmaxX));
        maxY = (fmaxY <= 0 ? 0 : (int) Math.ceil(fmaxY));

        return new Rectangle(0, 0, maxX, maxY);
    }

    /**
     * Creates an empty destination image with the correct size and number of
     * components.  A RasterFormatException may be thrown if the transformed
     * width or height is less than or equal to 0.
     * @param src       Source image for the filter operation
     * @param destCM    ColorModel of the destination.  If null, the
     *                  ColorModel of the source will be used.
     */
    public BufferedImage createCompatibleDestImage (BufferedImage src,
                                                    ColorModel destCM) {
        BufferedImage image;
        Rectangle r = getDestBounds(src).getBounds();
        int w = r.width;
        int h = r.height;

        if (destCM == null) {
            ColorModel cm = src.getColorModel();
            image = new BufferedImage(cm,
                          src.getRaster().createCompatibleWritableRaster(w,h),
                          cm.isAlphaPremultiplied());
        }
        else {
            image = new BufferedImage(destCM,
                            destCM.createCompatibleWritableRaster(w,h),
                            destCM.isAlphaPremultiplied());
        }

        return image;
    }
    
    /**
     * Creates an empty destination Raster with the correct size and number of
     * channels.
     */
    public WritableRaster createCompatibleDestRaster (Raster src) {
        Rectangle2D r = getDestBounds(src);
        return src.createCompatibleWritableRaster((int)r.getWidth(),
                                        (int)r.getHeight());
    }
    
    /**
     * Returns the location of the destination point given a
     * point in the source image.  If dstPt is non-null, it will
     * be used to hold the return value.
     */
    public Point2D getDestPoint (Point2D srcPt, Point2D dstPt) {
        return xform.transform (srcPt, dstPt);
    }
    
    /**
     * Returns the current transform.
     */
    public AffineTransform getTransform() {
        return xform;
    }

}
