/*
 * @(#)BandCombineOp.java	1.30 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.GraphicsEnvironment;
import java.awt.color.ICC_Profile;
import java.awt.geom.Rectangle2D;
import java.awt.Rectangle;
import java.awt.geom.Point2D;

/**
 * This class performs an arbitrary linear combination of bands 
 * in a Raster, using a specified matrix.  
 * <p>
 * For operations with a Raster, the width of the matrix must be equal to 
 * the number of bands, optionally plus one.  If there is one more column
 * in the matrix than the number of bands, there is an implied 1 at the
 * end of the pixel vector.  The height of the matrix must be equal to
 * the number of bands in the destination.  For example, a 3-banded Raster
 * might have the following transformation applied to
 * each pixel in order to invert the second band of the Raster.
 * <pre>
 *   [ 1.0   0.0   0.0    0.0  ]     [ b1 ]    
 *   [ 0.0  -1.0   0.0  255.0  ]  x  [ b2 ]
 *   [ 0.0   0.0   1.0    0.0  ]     [ b3 ]
 *                                   [ 1 ]
 * </pre>
 *
 * <p>
 * Note that the source and destination can be the same object.
 * @version 10 Feb 1997
 */
public class BandCombineOp implements  RasterOp {
    float[][] matrix;
    int nrows = 0;
    int ncols = 0;
    
    /**
     * This constructs a BandCombineOp with the specified matrix.
     * See the class comments for restrictions on the size of the
     * matrix.  The first subscript is the row index and the second
     * is the column index.
     */
    public BandCombineOp (float[][] matrix) {
        nrows = matrix.length;
        ncols = matrix[0].length;
        this.matrix = new float[nrows+1][ncols+1];
        for (int i=0; i < nrows; i++) {
            System.arraycopy(matrix[i], 0, this.matrix[i], 0, ncols);
        }
    }

    /**
     * Returns the matrix.
     */
    public float[][] getMatrix() {
        return (float[][]) matrix.clone();
    }
    
    /**
     * Transforms the Raster using the matrix in the constructor.
     * The IllegalArgumentException may be thrown if the number of
     * bands in the source or destination is incompatible with the
     * matrix.  See the class comments for more details.  If
     * the destination
     * is null, it will be created with a number of bands equalling
     * the number of rows in the matrix.  No exception will be thrown
     * if the operation causes a data overflow.
     */
    public WritableRaster filter(Raster src, WritableRaster dst) {
        int nBands = src.getNumBands();
        if (ncols != nBands && ncols != (nBands+1)) {
            throw new IllegalArgumentException("Number of columns in the "+
                                               "matrix ("+ncols+
                                               ") must be equal to the number"+
                                               " of bands ([+1]) in src ("+
                                               nBands+").");
        }
        if (dst == null) {
            dst = createCompatibleDestRaster(src);
        }
        else if (nrows != dst.getNumBands()) {
            throw new IllegalArgumentException("Number of rows in the "+
                                               "matrix ("+nrows+
                                               ") must be equal to the number"+
                                               " of bands ([+1]) in dst ("+
                                               nBands+").");
        }

        GraphicsEnvironment ge =
            GraphicsEnvironment.getLocalGraphicsEnvironment();
        ImagingLib imlib = ge.getImagingLib();

        if (imlib.filter(this, src, dst) != null) {
            return dst;
        }

        int[] pixel = null;
        int[] dstPixel = new int[dst.getNumBands()];
        float accum;
        int sminX = src.getMinX();
        int sY = src.getMinY();
        int dminX = dst.getMinX();
        int dY = dst.getMinY();
        int sX;
        int dX;
        if (ncols == nBands) {
            for (int y=0; y < src.getHeight(); y++, sY++, dY++) {
                dX = dminX;
                sX = sminX;
                for (int x=0; x < src.getWidth(); x++, sX++, dX++) {
                    pixel = src.getPixel(sX, sY, pixel);
                    for (int r=0; r < nrows; r++) {
                        accum = 0.f;
                        for (int c=0; c < ncols; c++) {
                            accum += matrix[r][c]*pixel[c];
                        }
                        dstPixel[r] = (int) accum;
                    }
                    dst.setPixel(dX, dY, dstPixel);
                }
            }
        }
        else {
            // Need to add constant
            for (int y=0; y < src.getHeight(); y++, sY++, dY++) {
                dX = dminX;
                sX = sminX;
                for (int x=0; x < src.getWidth(); x++, sX++, dX++) {
                    pixel = src.getPixel(sX, sY, pixel);
                    for (int r=0; r < nrows; r++) {
                        accum = 0.f;
                        for (int c=0; c < nBands; c++) {
                            accum += matrix[r][c]*pixel[c];
                        }
                        dstPixel[r] = (int) (accum+matrix[r][nBands]);
                    }
                    dst.setPixel(dX, dY, dstPixel);
                }
            }
        }

        return dst;
    }

    /**
     * Returns the bounding box of the transformed destination.  Since
     * this is not a geometric operation, the bounding box does not
     * change.
     * The IllegalArgumentException may be thrown if the number of
     * components in the source is incompatible with the matrix.  See
     * the class comments for more details.
     */
    public Rectangle2D getDestBounds (BufferedImage src) {
        return new Rectangle (0, 0,
			      src.getWidth(), src.getHeight());
    }

    /**
     * Returns the bounding box of the transformed destination.  Since
     * this is not a geometric operation, the bounding box does not
     * change.
     * The IllegalArgumentException may be thrown if the number of
     * bands in the source is incompatible with the matrix.  See
     * the class comments for more details.
     */
    public Rectangle2D getDestBounds (Raster src) {
	//        return new Rectangle (0, 0,
        //                      src.getWidth(), src.getHeight());
	return src.getBounds();
    }

    
    /**
     * Creates an empty destination Raster with the correct size and
     * number of bands.
     * The IllegalArgumentException may be thrown if the number of
     * bands in the source is incompatible with the matrix or if the
     * number of rows in the matrix is not equal to the number of
     * bands in the source image.  See
     * the class comments for more details.
     * @param src       Source Raster for the filter operation
     */
    public WritableRaster createCompatibleDestRaster (Raster src) {
        int nBands = src.getNumBands();
        if ((ncols != nBands) && (ncols != (nBands+1))) {
            throw new IllegalArgumentException("Number of columns in the "+
                                               "matrix ("+ncols+
                                               ") must be equal to the number"+
                                               " of bands ([+1]) in src ("+
                                               nBands+").");
        }
        if (src.getNumBands() == nrows) {
            return src.createCompatibleWritableRaster();
        }
        else {
            throw new IllegalArgumentException("Don't know how to create a "+
                                               " compatible Raster with "+
                                               nrows+" bands.");
        }
    }

    /**
     * Returns the location of the destination point given a
     * point in the source image.  If dstPt is non-null, it will
     * be used to hold the return value.  Since this is not a geometric
     * operation, the srcPt will equal the dstPt.
     */
    public Point2D getDestPoint (Point2D srcPt, Point2D dstPt) {
        if (dstPt == null) {
            dstPt = new Point2D.Float();
        }
	dstPt.setLocation(srcPt.getX(), srcPt.getY());

        return dstPt;
    }
    
}
