/*
 * @(#)BufferedImageFilter.java	1.17 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.util.Hashtable;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageFilter;

/**
 * This class subclasses an ImageFilter to provide a simple means of
 * using a single-source/single-destination image operator
 * (BufferedImageOp) to filter
 * a BufferedImage or Raster in the Image Producer/Consumer/Observer
 * paradigm. Examples of these image operators are: ConvolveOp,
 * AffineTransformOp and LookupOp.
 *
 * @see ImageFilter
 * @see BufferedImage
 * @see Raster
 * @see WritableRaster
 * @see BufferedImageOp
 * @version 10 Feb 1997
 */

public class BufferedImageFilter extends ImageFilter
implements RasterImageConsumer, Cloneable {
    private BufferedImageOp bufferedImageOp;
    
    /**
     * Constructs a BufferedImageFilter with the
     * specified single-source/single-destination operator.
     */
    public BufferedImageFilter (BufferedImageOp op) {
        super();
        bufferedImageOp = op;
    }
    
    /**
     * The consumer of the particular image data stream for which this
     * instance of the ImageFilter is filtering data.  It is not
     * initialized during the constructor, but rather during the
     * getFilterInstance() method call when the FilteredImageSource
     * is creating a unique instance of this object for a particular
     * image data stream.
     * @see #getFilterInstance
     * @see ImageConsumer
     */
    private ImageConsumer consumer;

    /**
     * Returns the BufferedImageOp.
     */
    public BufferedImageOp getBufferedImageOp() {
        return bufferedImageOp;
    }
    
    /**
     * Returns a unique instance of a BufferedImageFilter object which will
     * actually perform the filtering for the specified ImageConsumer.
     */
    public ImageFilter getFilterInstance(ImageConsumer ic) {
        BufferedImageFilter instance = (BufferedImageFilter) clone();
        instance.consumer = ic;
        return instance;
    }

    /**
     * Filters the information provided in the setPixels method of the
     * RasterImageConsumer interface which takes a WritableRaster.
     * The filter used will be the BufferedImageOp that was passed
     * to the constructor.
     * @see ImageConsumer#setPixels
     */
    public void setPixels(int x, int y,
			  ColorModel model, WritableRaster raster) {
        if (consumer instanceof RasterImageConsumer) {
            RasterImageConsumer rconsumer = (RasterImageConsumer) consumer;
            
            rconsumer.setPixels (x, y, model, raster);
        }
        else {
            int w = raster.getWidth();
            int h = raster.getHeight();
            Object obj = raster.getPixelData(raster.getMinX(),
					     raster.getMinY(),
					     w, h, null);
            int data[];
            if (obj instanceof int[]) { 
                data = (int[])obj;
            } else {
                data = DataBuffer.toIntArray(obj);
            }
            consumer.setPixels(x, y, w, h, model, data, 0,
                               w*raster.getNumDataElements());
        }
    }

}
