/*
 * @(#)ComponentColorModel.java	1.38 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.color.ColorSpace;

/**
 * A ColorModel class that can handle an arbitrary ColorSpace and
 * an array of color components to match the ColorSpace.  This
 * class can be used to represent most color models on most
 * types of GraphicsDevices.  
 *
 * @see ColorModel
 *
 * @version 10 Feb 1997
 */
public class ComponentColorModel extends ColorModel {
    
    /**
     * Constructs a ComponentColorModel from the specified parameters.
     * The bits array specifies the number of significant bits per
     * component in the color model (color components and alpha).  An
     * IllegalArgumentException will be thrown if the length of the
     * array does not match the number of components.
     * If hasAlpha is true, then the boolean, isAlphaPremultiplied, specifies
     * how to interpret color data that are passed to this ColorModel.  If
     * it is true, the color data will have been multiplied with alpha.
     * @param colorSpace       ColorSpace associated with this color model
     * @param bits             Number of significant bits per component
     * @param hasAlpha         If true, this color model supports alpha.
     * @param isAlphaPremultiplied If true, alpha is premultiplied.
     * @param transparency     Specifies what alpha values can be represented
     *                         by this color model.
     * @see ColorSpace
     * @see java.awt.Transparency
     */
    public ComponentColorModel (ColorSpace colorSpace,
                                int[] bits,
                                boolean hasAlpha,
                                boolean isAlphaPremultiplied,
                                int transparency) {
        super (bits, colorSpace, hasAlpha, isAlphaPremultiplied, transparency);
    }

    /**
     * Returns the red
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.
     */
    public int getRed(int pixel) {
        if (numComponents > 1) {
            throw new
                IllegalArgumentException("More than one component per pixel");
        }

        // Normalize the pixel in order to convert it
        float[] norm = { (float) pixel / ((1<<nBits[0]) - 1) };
        float[] rgb = colorSpace.toRGB(norm);

        return (int) rgb[0]*255;
    }

    /**
     * Returns the green
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.
     */
    public int getGreen(int pixel) {
        if (numComponents > 1) {
            throw new
                IllegalArgumentException("More than one component per pixel");
        }

        // Normalize the pixel in order to convert it
        float[] norm = { (float) pixel / ((1<<nBits[0]) - 1) };
        float[] rgb = colorSpace.toRGB(norm);
        
        return (int) rgb[1]*255;
    }

    /**
     * Returns the blue
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.
     */
    public int getBlue(int pixel) {
        if (numComponents > 1) {
            throw new
                IllegalArgumentException("More than one component per pixel");
        }

        // Normalize the pixel in order to convert it
        float[] norm = { (float) pixel / ((1<<nBits[0]) - 1) };
        float[] rgb = colorSpace.toRGB(norm);
        
        return (int) rgb[2]*255;
    }

    /**
     * Returns the alpha
     * color component for the specified pixel, scaled from 0 to 255
     * in the default RGB ColorSpace, sRGB.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.
     */
    public int getAlpha(int pixel) {
        if (supportsAlpha == false) {
            return 255;
        }
        if (numComponents > 1) {
            throw new
                IllegalArgumentException("More than one component per pixel");
        }

        return (pixel/((1<<nBits[0])-1))*255;
    }

    /**
     * Returns the color of the pixel in the default RGB color model.
     * An IllegalArgumentException will be thrown if there is more than
     * one component in this ColorModel.
     * This method will throw an IllegalArgumentException since this
     * ColorModel is unable to extract alpha and color components from
     * a single integer pixel.
     */
    public int getRGB(int pixel) {
        if (numComponents > 1) {
            throw new
                IllegalArgumentException("More than one component per pixel");
        }

	return (getAlpha(pixel) << 24)
	    | (getRed(pixel) << 16)
	    | (getGreen(pixel) << 8)
	    | (getBlue(pixel) << 0);
    }


    /**
     * Returns the red
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getRed(Object inData) {
        // REMIND: Should this be CS_sRGB or TYPE_RGB?
        if (colorSpaceType == ColorSpace.TYPE_RGB) {
            boolean needAlpha = (supportsAlpha && isAlphaPremultiplied);
            int alp = 0;
            int red = 0;
            switch (transferType) {
                case DataBuffer.BYTE_DATA:
                   byte bdata[] = (byte[])inData;
                   red = bdata[0] & 0xff;
                   if (needAlpha) {
                       alp = bdata[numColorComponents];
                   }
                break;
                case DataBuffer.SHORT_DATA:
                   short sdata[] = (short[])inData;
                   red = sdata[0]&0xffff;
                   if (needAlpha) {
                       alp = sdata[numColorComponents];
                   }
                break;
                case DataBuffer.INT_DATA:
                   int idata[] = (int[])inData;
                   red = idata[0];
                   if (needAlpha) {
                       alp = idata[numColorComponents];
                   }
                break;
                default:
                   throw new NoSuchMethodError("This method has not been "+
                       "implemented for transferType " + transferType);
            }
            if (nBits[0] != 8) {
                int shift = nBits[0] - 8;
                red = ((shift > 0) 
                       ? (red>>shift)
                       : (red<<(-shift)));
            }
            if (needAlpha) {
                return (int) (red/(1<<nBits[numColorComponents])-1.f);
            }
            else {
                return red;
            }
        }
        else if (colorSpaceType == ColorSpace.TYPE_GRAY) {
            return getGray(inData);
        }

        int pixel[];
        if (inData instanceof int[]) {
            pixel = (int[])inData;
        } else {
            pixel = DataBuffer.toIntArray(inData);
            if (pixel == null) {
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
            }
        }

        // Normalize the pixel in order to convert it
        float[] norm = getNormalizedComponents(pixel, 0, null, 0);
        float[] rgb = colorSpace.toRGB(norm);
        int color;
        
        if (isAlphaPremultiplied && supportsAlpha) {
            color = (int) (rgb[0]*255*getNormAlpha(pixel));
        }
        else {
            color = (int) rgb[0]*255;
        }
        
        return color;
    }
   

    /**
     * Returns the green
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getGreen(Object inData) {
        // REMIND: Should this be CS_sRGB or TYPE_RGB?
        if (colorSpaceType == ColorSpace.TYPE_RGB) {
            int green = 0;
            switch (transferType) {
                case DataBuffer.BYTE_DATA:
                   byte bdata[] = (byte[])inData;
                   green = bdata[1] & 0xff;
                break;
                case DataBuffer.SHORT_DATA:
                   short sdata[] = (short[])inData;
                   green = sdata[1];
                break;
                case DataBuffer.INT_DATA:
                   int idata[] = (int[])inData;
                   green = idata[1];
                break;
                default:
                   throw new NoSuchMethodError("This method has not been "+
                       "implemented for transferType " + transferType);
            }
            if (nBits[1] == 8) {
                return green;
            }
            else {
                int shift = nBits[1] - 8;
                return ((shift > 0) 
                        ? (green>>shift)
                        : (green<<(-shift)));

            }
        }
        else if (colorSpaceType == ColorSpace.TYPE_GRAY) {
            return getGray(inData);
        }

        int pixel[];
        if (inData instanceof int[]) {
            pixel = (int[])inData;
        } else {
            pixel = DataBuffer.toIntArray(inData);
            if (pixel == null) {
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
            }
        }

        // Normalize the pixel in order to convert it
        float[] norm = getNormalizedComponents(pixel, 0, null, 0);
        float[] rgb = colorSpace.toRGB(norm);
        int color;
        
        if (isAlphaPremultiplied && supportsAlpha) {
            color = (int) (rgb[1]*255*getNormAlpha(pixel));
        }
        else {
            color = (int) rgb[1]*255;
        }
        
        return color;
    }
   

    /**
     * Returns the blue
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getBlue(Object inData) {
        // REMIND: Should this be CS_sRGB or TYPE_RGB?
        if (colorSpaceType == ColorSpace.TYPE_RGB) {
            int blue = 0;
            switch (transferType) {
                case DataBuffer.BYTE_DATA:
                   byte bdata[] = (byte[])inData;
                   blue = bdata[2] & 0xff;
                break;
                case DataBuffer.SHORT_DATA:
                   short sdata[] = (short[])inData;
                   blue = sdata[2];
                break;
                case DataBuffer.INT_DATA:
                   int idata[] = (int[])inData;
                   blue = idata[2];
                break;
                default:
                   throw new NoSuchMethodError("This method has not been "+
                       "implemented for transferType " + transferType);
            }
            if (nBits[2] == 8) {
                return blue;
            }
            else {
                int shift = nBits[2] - 8;
                return ((shift > 0) 
                        ? (blue>>shift)
                        : (blue<<(-shift)));
            }
        }
        else if (colorSpaceType == ColorSpace.TYPE_GRAY) {
            return getGray(inData);
        }

        int pixel[];
        if (inData instanceof int[]) {
            pixel = (int[])inData;
        } else {
            pixel = DataBuffer.toIntArray(inData);
            if (pixel == null) {
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
            }
        }

        // Normalize the pixel in order to convert it
        float[] norm = getNormalizedComponents(pixel, 0, null, 0);
        float[] rgb = colorSpace.toRGB(norm);
        int color;
        
        if (isAlphaPremultiplied && supportsAlpha) {
            color = (int) (rgb[2]*255*getNormAlpha(pixel));
        }
        else {
            color = (int) rgb[2]*255;
        }
        
        return color;
    }

    /**
     * Returns the alpha component for the specified pixel, scaled from
     * 0 to 255. The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     */
    public int getAlpha(Object inData) {
        if (supportsAlpha == false) {
            return 255;
        }

        int alpha = 0;
        int aIdx = numColorComponents;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               alpha = bdata[aIdx] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               alpha = sdata[aIdx];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               alpha = idata[aIdx];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }

        if (nBits[aIdx] == 8) {
            return alpha;
        }
        else {
            int shift = nBits[aIdx] - 8;
            return ((shift > 0) 
                    ? (alpha>>shift)
                    : (alpha<<(-shift)));
        }
    }

    // This method returns a gray color that can be mapped to
    // RGB values in getRGB().  It assumes that the colorspace is TYPE_GRAY.
    private int getGray(Object inData) {
        boolean needAlpha = (supportsAlpha && isAlphaPremultiplied);
        int alp = 0;
        int gray;
        switch (transferType) {
        case DataBuffer.BYTE_DATA:
            byte bdata[] = (byte[])inData;
            gray = bdata[0] & 0xff;
            if (needAlpha) {
                alp = bdata[1]&0xff;
            }
            break;
        case DataBuffer.SHORT_DATA:
            short sdata[] = (short[])inData;
            gray = sdata[0] & 0xffff;
            if (needAlpha) {
                alp = sdata[1]&0xff;
            }
            break;
        case DataBuffer.INT_DATA:
            int idata[] = (int[])inData;
            gray = idata[0];
            if (needAlpha) {
                alp = idata[1]&0xff;
            }
            break;
        default:
            throw new NoSuchMethodError("This method has not been "+
                                        "implemented for transferType " + transferType);
        }

        if (nBits[0] != 8) {
            int shift = nBits[0] - 8;
            gray = ((shift > 0)
                    ? (gray>>shift)
                    : (gray<<(-shift)));
        }
        return (!needAlpha
                ? gray
                : (int)(gray/(1<<nBits[numColorComponents])-1.f)
                   );

    }
    
    private float getNormAlpha(int pixel[]) {
        return (float) (pixel[numColorComponents] /
                        ((1<<nBits[numColorComponents])-1.f));
    }
    
    /**
     * Returns a pixel in the default RGB color model given a pixel array
     * in the current color model.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     * @see ColorModel#getRGBdefault
     */
    public int getRGB(Object inData) {
        if (colorSpaceType == ColorSpace.TYPE_GRAY) {
            int gray = getGray(inData);
            return (getAlpha(inData) << 24)
                | (gray << 16)
                | (gray <<  8)
                | gray;
        }
        return (getAlpha(inData) << 24)
            | (getRed(inData) << 16)
            | (getGreen(inData) << 8)
            | (getBlue(inData));
    }

    /**
     * Returns a pixel array given a pixel in the default RGB color model.
     * The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * This pixel array can then be passed to the setPixelData method of
     * a Raster object.  If the pixel variable is null, a new array will
     * be allocated.  The pixel array will be returned.
     */
    public Object getPixelData(int rgb, Object inData) {
        //REMIND: maybe more efficient not to use int array for
        //DataBuffer.SHORT_DATA and DataBuffer.INT_DATA
        int pixel[] = null;
        if (transferType == DataBuffer.INT_DATA &&
            inData != null) {
           pixel = (int[])inData;
        } else {
            pixel = new int[numComponents];
        }
        // REMIND: Use rendering hints?
        if (! is_sRGB) {
            // Need to convert the color
            float[] norm = new float[3];
            norm[0] = ((rgb>>16)&0xff)/255.f;
            norm[1] = ((rgb>>8)&0xff)/255.f;
            norm[2] = ((rgb>>0)&0xff)/255.f;

            norm = colorSpace.fromRGB(norm);
            for (int i = 0; i < numColorComponents; i++) {
                pixel[i] = (int)(norm[i]*((1<<nBits[i]) - 1));
            }
        }
        else {
            int alp = (rgb>>24)&0xff;
            int red = (rgb>>16)&0xff;
            int grn = (rgb>>8) &0xff;
            int blu = (rgb)    &0xff;
            if (isAlphaPremultiplied) {
                float norm = alp/255.f;
                if (nBits[0] == 8) {
                    pixel[0] = (int)(red*norm);
                }
                else {
                    pixel[0] = (int)((red*norm)/255.f * ((1<<nBits[0])-1));
                }
                if (nBits[1] == 8) {
                    pixel[1] = (int)(grn*norm);
                }
                else {
                    pixel[1] = (int)((grn*norm)/255.f * ((1<<nBits[1])-1));
                }
                if (nBits[2] == 8) {
                    pixel[2] = (int)(blu*norm);
                }
                else {
                    pixel[2] = (int)((blu*norm)/255.f * ((1<<nBits[2])-1));
                }
                if (supportsAlpha) {
                    if (nBits[3] == 8) {
                        pixel[3] = alp;
                    }
                    else {
                        pixel[3] = (int)(norm * ((1<<nBits[3])-1));
                    }
                }
            }
            else {
                if (nBits[0] == 8) {
                    pixel[0] = red;
                }
                else {
                    pixel[0] = (int)(red/255.f * ((1<<nBits[0])-1));
                }
                if (nBits[1] == 8) {
                    pixel[1] = (int)(grn);
                }
                else {
                    pixel[1] = (int)(grn/255.f * ((1<<nBits[1])-1));
                }
                if (nBits[2] == 8) {
                    pixel[2] = (int)(blu);
                }
                else {
                    pixel[2] = (int)(blu/255.f * ((1<<nBits[2])-1));
                }
                if (supportsAlpha) {
                    if (nBits[3] == 8) {
                        pixel[3] = alp;
                    }
                    else {
                        pixel[3] = (int)(alp/255.f * ((1<<nBits[3])-1));
                    }
                }
            }
        }
        
        switch (transferType) {
            case DataBuffer.BYTE_DATA: {
               byte bdata[];
               if (inData == null) {
                   bdata = new byte[numComponents];
               } else {
                   bdata = (byte[])inData;
               }
               for (int i = 0; i < numComponents; i++) {
                   bdata[i] = (byte)(0xff&pixel[i]);
               }
               return bdata;
            }
            case DataBuffer.SHORT_DATA:{
               short sdata[];
               if (inData == null) {
                   sdata = new short[numComponents];
               } else {
                   sdata = (short[])inData;
               }
               for (int i = 0; i < numComponents; i++) {
                   sdata[i] = (short)(pixel[i]);
               }
               return sdata;
            }
            case DataBuffer.INT_DATA:
               return pixel;
        }
        throw new IllegalArgumentException("This method has not been "+
                 "implemented for transferType " + transferType);
    }
    
    /**
     * Returns an array of all of the components given a pixel 
     * in the current color model.  If the color model is not
     * an IndexColorModel or DirectColorModel, an IllegalArgumentException
     * will be thrown.  This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    public int[] getComponents(int pixel, int[] components, int offset) {
        if (numComponents > 1) {
            throw new
                IllegalArgumentException("More than one component per pixel");
        }
        if (components == null) {
            components = new int[offset+1];
        }

        components[offset+0] = (pixel & ((1<<nBits[0]) - 1));
        return components;
    }
    
    /**
     * Returns an array of all of the components given a pixel array
     * in the current color model.  The pixel array is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    public int[] getComponents(Object inPixel, int[] components, int offset) {
        int pixel[];
        if (inPixel instanceof int[]) {
            pixel = (int[])inPixel;
        } else {
            pixel = DataBuffer.toIntArray(inPixel);
            if (pixel == null) {
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
            }
        }
        if (pixel.length < numComponents) {
            throw new IllegalArgumentException
                ("Length of pixel array < number of components in model");
        }
        if (components == null) {
            components = new int[offset+numComponents];
        }
        else if ((components.length-offset) < numComponents) {
            throw new IllegalArgumentException
                ("Length of components array < number of components in model");
        }
        System.arraycopy(pixel, 0, components, offset, numComponents);

        return components;
    }
    
    /**
     * Returns a pixel, given a discrete components (unpacked color/alpha
     * values) in this color model.
     */
    public int setComponents(int[] components, int offset) {
        if (numComponents == 1) {
            return components[offset+0];
        }
        throw new IllegalArgumentException("This model returns "+
                                           numComponents+
                                           " elements in the pixel array.");
    }
    
    /**
     * Returns a pixel array, given a discrete components (unpacked
     * color/alpha values) in this color model.  The pixel is
     * described by a discrete array of data elements of type
     * transferType passed in as an object reference.
     * The returned array can be passed to the setPixelData methods of Raster.
     * objects.
     */
    public Object setComponents(int[] components, int offset, Object obj) {
        if ((components.length-offset) < numComponents) {
            throw new IllegalArgumentException("Component array too small"+
                                               " (should be "+numComponents);
        }
        switch(transferType) {
        case DataBuffer.INT_DATA:
            {
                int[] pixel;
                if (obj == null) {
                    pixel = new int[components.length];
                }
                else {
                    pixel = (int[]) obj;
                }
                System.arraycopy(components, offset, pixel, 0,
                                 numComponents);
                return pixel;
            }
        
        case DataBuffer.BYTE_DATA:
            {
                byte[] pixel;
                if (obj == null) {
                    pixel = new byte[components.length];
                }
                else {
                    pixel = (byte[]) obj;
                }
                for (int i=0; i < numComponents; i++) {
                    pixel[i] = (byte) components[offset+i];
                }
                return pixel;
            }
        
        case DataBuffer.SHORT_DATA:
            {
                short[] pixel;
                if (obj == null) {
                    pixel = new short[components.length];
                }
                else {
                    pixel = (short[]) obj;
                }
                for (int i=0; i < numComponents; i++) {
                    pixel[i] = (short) components[offset+i];
                }
                return pixel;
            }
        
        default:
            throw new NoSuchMethodError("This method has not been "+
                                        "implemented for transferType " +
                                        transferType);
        }
    }

    /**
     * Forces the Raster data to match the state specified in the
     * isAlphaPremultiplied variable, assuming the data is currently
     * correctly described by this ColorModel.  It may multiply or divide
     * the color Raster data by alpha, or do nothing if the data is
     * in the correct state.  If the data needs to be coerced, this
     * method will also return an instance of this ColorModel with
     * the isAlphaPremultiplied flag set appropriately.
     */
    public ColorModel coerceData (WritableRaster raster, 
                                  boolean isAlphaPremultiplied) {
        if ((supportsAlpha == false) ||
            (this.isAlphaPremultiplied == isAlphaPremultiplied))
        {
            // Nothing to do
            return this;
        }
        
        int w = raster.getWidth();
        int h = raster.getHeight();
        int aIdx = raster.getNumBands() - 1;
        int alpha;
        int rminX = raster.getMinX();
        int rY = raster.getMinY();
        int rX;
        if (isAlphaPremultiplied) {
            switch (transferType) {
                case DataBuffer.BYTE_DATA: {
                    byte pixel[] = null;
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = (byte[])raster.getPixelData(rX, rY, pixel);
                            alpha = pixel[aIdx];
                            if (alpha != 0) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] *= alpha;
                                }
                                raster.setPixelData(rX, rY, pixel);
                            }
                        }
                    }
                }
                break;
                case DataBuffer.SHORT_DATA: {
                    short pixel[] = null;
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = (short[])raster.getPixelData(rX, rY,
                                                                 pixel);
                            alpha = pixel[aIdx];
                            if (alpha != 0) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] *= alpha;
                                }
                                raster.setPixelData(rX, rY, pixel);
                            }
                        }
                    }
                }
                break;
                case DataBuffer.INT_DATA: {
                    int pixel[] = null;
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = (int[])raster.getPixelData(rX, rY, pixel);
                            alpha = pixel[aIdx];
                            if (alpha != 0) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] *= alpha;
                                }
                                raster.setPixelData(rX, rY, pixel);
                            }
                        }
                    }
                }
                break;
                default:
                    throw new NoSuchMethodError("This method has not been "+
                         "implemented for transferType " + transferType);
            }
        }
        else {
            // We are premultiplied and want to divide it out
            switch (transferType) {
                case DataBuffer.BYTE_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            byte pixel[] = null;
                            pixel = (byte[])raster.getPixelData(rX, rY, pixel);
                            alpha = pixel[aIdx];
                            for (int c=0; c < aIdx; c++) {
                                if (alpha != 0) {
                                    pixel[c] /= alpha;
                                }
                            }
                            raster.setPixelData(rX, rY, pixel);
                        }
                    }
                }
                break;
                case DataBuffer.SHORT_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            short pixel[] = null;
                            pixel = (short[])raster.getPixelData(rX, rY,
                                                                 pixel);
                            alpha = pixel[aIdx];
                            for (int c=0; c < aIdx; c++) {
                                if (alpha != 0) {
                                    pixel[c] /= alpha;
                                }
                            }
                            raster.setPixelData(rX, rY, pixel);
                        }
                    }
                }
                break;
                case DataBuffer.INT_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            int pixel[] = null;
                            pixel = (int[])raster.getPixelData(rX, rY, pixel);
                            alpha = pixel[aIdx];
                            for (int c=0; c < aIdx; c++) {
                                if (alpha != 0) {
                                    pixel[c] /= alpha;
                                }
                            }
                            raster.setPixelData(rX, rY, pixel);
                        }
                    }
                }
                break;
                default:
                    throw new NoSuchMethodError("This method has not been "+
                         "implemented for transferType " + transferType);
            }
        }

        // Return a new color model
        return new ComponentColorModel(colorSpace, nBits, supportsAlpha,
                                       isAlphaPremultiplied, transparency);

    }

    /**
      * Returns true if raster is compatible with this ColorModel and
      * false if it is not.
      */
    public boolean isCompatibleRaster(Raster raster) {

	boolean flag = true;
        SampleModel sm = raster.getSampleModel();
        ComponentSampleModel csm;
        if (sm instanceof ComponentSampleModel) {
	csm = (ComponentSampleModel) sm;
        }
        else {
            return false;
        }
        if (csm.getNumBands() != getNumComponents()) {
            return false;
        }
	for (int i=0; i<nBits.length; i++) {
	    if (csm.getSampleSize(i) != nBits[i])
		flag = false;
	}
        return ( (raster.getTransferType() == transferType) && flag);
    }
    
    /**
     * Creates a WritableRaster with the specified width and height, that 
     * has a channel layout compatible with this ColorModel.  
     * @see WritableRaster
     */
    public WritableRaster createCompatibleWritableRaster (int w, int h) {
        int[] channelOffsets  = new int[numComponents];
        int dataSize = w*h*numComponents;
        WritableRaster raster = null;

        // Load the channel parameters
        for (int i=0; i < numComponents; i++) {
            channelOffsets[i] = i;
        }
       
 
        // Create the raster
        if (maxBits <= 8) {
            raster = Raster.createComponentRaster(DataBuffer.BYTE_DATA,
                                                  w,h,w*numComponents,
                                                  numComponents, 
                                                  channelOffsets,null);
        }
        else if (maxBits <= 16) {
            raster = Raster.createComponentRaster(DataBuffer.SHORT_DATA,
                                                  w,h,w*numComponents,
                                                  numComponents,  
                                                  channelOffsets,null);
        }
        else {
            raster = Raster.createComponentRaster(DataBuffer.INT_DATA,
                                                  w,h,w*numComponents,
                                                  numComponents,  
                                                  channelOffsets,null);
        }


        return raster;
    }

    /**
     * Compares this color model with another for equality.  
     */
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }

        if (obj.getClass() !=  getClass()) {
            return false;
        }
        
        return true;
    }

}

