/*
 * @(#)DirectColorModel.java	1.50 98/03/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.color.ColorSpace;
import java.awt.Transparency;

/**
 * A ColorModel class that represents pixel values
 * that have RGB color components embedded directly in the bits of the
 * pixel itself.  This color model is similar to an X11 TrueColor visual.  
 * The default RGB ColorModel specified by the ColorModel.getRGBdefault
 * method is a DirectColorModel with the following parameters:
 * <pre>
 * Number of bits:        32
 * Red mask:              0x00ff0000
 * Green mask:            0x0000ff00
 * Blue mask:             0x000000ff
 * Alpha mask:            0xff000000
 * Color space:           sRGB
 * isAlphaPremultiplied:  False
 * </pre>
 *
 * <p>Many of the methods in this class are final. This is because the
 * underlying native graphics code makes  assumptions about the layout
 * and operation of this class  and those assumptions are reflected in
 * the implementations of the methods here that are marked final.  You
 * can subclass this class  for other reasons,  but you cannot override
 * or modify the behaviour of those methods.
 *
 * @see ColorModel
 * @see ColorSpace
 *
 * @version 10 Feb 1997
 */
public class DirectColorModel extends PackedColorModel {
    private int red_mask;
    private int green_mask;
    private int blue_mask;
    private int alpha_mask;
    private int red_offset;
    private int green_offset;
    private int blue_offset;
    private int alpha_offset;
    private int red_scale;
    private int green_scale;
    private int blue_scale;
    private int alpha_scale;
    
    /**
     * Constructs a DirectColorModel from the given masks specifying
     * which bits in the pixel contain the red, green and blue color
     * components.  Pixels described by this color model will all
     * have alpha components of 255 (fully opaque).  All of the bits
     * in each mask must be contiguous and fit in the specified number
     * of least significant bits of the integer.
     */
    public DirectColorModel(int bits,
			    int rmask, int gmask, int bmask) {
	this(bits, rmask, gmask, bmask, 0);
    }

    /**
     * Constructs a DirectColorModel from the given masks specifying
     * which bits in the pixel contain the alpha, red, green and blue
     * color components.  All of the bits in each mask must be contiguous
     * and fit in the specified number of least significant bits of the
     * integer.  The ColorSpace will be the default sRGB space, and
     * alpha, if present, will not be premultiplied.
     */
    public DirectColorModel(int bits, int rmask, int gmask,
                            int bmask, int amask) {
        super (ColorSpace.getInstance(ColorSpace.CS_sRGB),
               bits, rmask, gmask, bmask, amask, false,
               amask == 0 ? Transparency.OPAQUE : Transparency.TRANSLUCENT);
        setFields();
    }

    /**
     * Constructs a DirectColorModel from the specified parameters.
     * The masks specify
     * which bits in the pixel contain the alpha, red, green and blue
     * color components.  All of the bits in each mask must be contiguous
     * and fit in the specified number of least significant bits of the
     * integer.  The ColorSpace of the color model is specified by
     * space and must be of ColorSpace.TYPE_RGB.  If there is alpha, the
     * boolean,
     * isAlphaPremultiplied, specifies how to interpret color data that
     * are passed to this ColorModel.  If it is true, the color data will
     * have been multiplied with alpha.
     */
    public DirectColorModel(ColorSpace space, int bits, int rmask,
                            int gmask, int bmask, int amask,
                            boolean isAlphaPremultiplied) {
        super (space, bits, rmask, gmask, bmask, amask,
               isAlphaPremultiplied,
               amask == 0 ? Transparency.OPAQUE : Transparency.TRANSLUCENT);  
        setFields();
    }

    /**
     * Returns the mask indicating which bits in a pixel contain the red
     * color component.
     */
    final public int getRedMask() {
	return maskArray[0];
    }

    /**
     * Returns the mask indicating which bits in a pixel contain the green
     * color component.
     */
    final public int getGreenMask() {
	return maskArray[1];
    }

    /**
     * Returns the mask indicating which bits in a pixel contain the blue
     * color component.
     */
    final public int getBlueMask() {
	return maskArray[2];
    }

    /**
     * Returns the mask indicating which bits in a pixel contain the alpha
     * transparency component.
     */
    final public int getAlphaMask() {
        if (supportsAlpha) {
            return maskArray[3];
        } else {
            return 0;
        }
    }


    /**
     * Returns the red color component for the specified pixel in the
     * range 0-255 in the default RGB ColorSpace, sRGB.  
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    final public int getRed(int pixel) {
	int r = ((pixel & maskArray[0]) >>> maskOffsets[0]);
	if (scaleFactors[0] != 1.) {
	    r = (int)(r * scaleFactors[0]);
	}
        if (isAlphaPremultiplied) {
            int a = getAlpha(pixel);
            r = (a == 0) ? 0 : (r * 255/a);
        }
	return r;
    }

    /**
     * Returns the green color component for the specified pixel in the
     * range 0-255 in the default RGB ColorSpace, sRGB.  The
     * returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    final public int getGreen(int pixel) {
	int g = ((pixel & maskArray[1]) >>> maskOffsets[1]);
	if (scaleFactors[1] != 1.) {
	    g = (int) (g * scaleFactors[1]);
	}

        if (isAlphaPremultiplied) {
            int a = getAlpha(pixel);
            g = (a == 0) ? 0 : (g * 255/a);
        }
	return g;
    }

    /**
     * Returns the blue color component for the specified pixel in the
     * range 0-255 in the default RGB ColorSpace, sRGB.  The
     * returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    final public int getBlue(int pixel) {
	int b = ((pixel & maskArray[2]) >>> maskOffsets[2]);
	if (scaleFactors[2] != 1.) {
	    b = (int)(b * scaleFactors[2]);
	}

        if (isAlphaPremultiplied) {
            int a = getAlpha(pixel);
            b = (a == 0) ? 0 : (b * 255/a);
        }
	return b;
    }

    /**
     * Return the alpha transparency value for the specified pixel in the
     * range 0-255.
     */
    final public int getAlpha(int pixel) {
	if (!supportsAlpha) return 255;
	int a = ((pixel & maskArray[3]) >>> maskOffsets[3]);
	if (scaleFactors[3] != 1.) {
	    a = (int)(a * scaleFactors[3]);
	}
	return a;
    }

    /**
     * Returns the color of the pixel in the default RGB color model.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).
     * @see ColorModel#getRGBdefault
     */
    final public int getRGB(int pixel) {
	return (getAlpha(pixel) << 24)
	    | (getRed(pixel) << 16)
	    | (getGreen(pixel) << 8)
	    | (getBlue(pixel) << 0);
    }
    
    /**
     * Returns the red
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getRed(Object inData) {
        int pixel=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getRed(pixel);
    }
   

    /**
     * Returns the green
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getGreen(Object inData) {
        int pixel=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getGreen(pixel);
    }
   

    /**
     * Returns the blue
     * color component for the specified pixel, scaled from 0 to 255 in
     * the default RGB ColorSpace, sRGB.  The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    public int getBlue(Object inData) {
        int pixel=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getBlue(pixel);
    }

    /**
     * Returns the alpha color component for the specified pixel,
     * scaled from 0 to 255.
     * The pixel is
     * described by a discrete array of data elements of type
     * transferType passed in as an object reference.
     */
    public int getAlpha(Object inData) {
        int pixel=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getAlpha(pixel);
    }

    /**
     * Returns a pixel in the default RGB color model given a pixel 
     * in the current color model.
     * The pixel is described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     * @see ColorModel#getRGBdefault
     */
    public int getRGB(Object inData) {
        int pixel=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getRGB(pixel);
    }

    /**
     * Returns a pixel array of type transferType by object reference
     * given a pixel in the default RGB color model.
     * The pixel is
     * described by a discrete array of data elements of type
     * transferType passed in as an object reference.
     * This pixel array can then be passed to the setPixelData method of
     * a Raster object.  If the pixel variable is null, a new array will
     * be allocated.  The pixel array will be returned.
     */
    public Object getPixelData(int rgb, Object inData) {
        //REMIND: maybe more efficient not to use int array for
        //DataBuffer.SHORT_DATA and DataBuffer.INT_DATA
        int i;
        int pixel[] = null;
        if (transferType == DataBuffer.INT_DATA &&
            inData != null) {
           pixel = (int[])inData;
        } else {
            pixel = new int[1];
        }

        if (this == ColorModel.getRGBdefault()) {
            pixel[0] = rgb;
        } else {
            double normAlpha = 1.;
            int c;
            if (isAlphaPremultiplied) {
                normAlpha = ((rgb>>24)&0xff)/255.;
            }

            double value;
            int[] shifts = {16, 8, 0, 24};
            int nbands = 3+(supportsAlpha ? 1 : 0);

            pixel[0] = 0;
            for (i=0; i < nbands; i++) {
                c = (int)(((rgb>>shifts[i])&0xff)*normAlpha/scaleFactors[i]);
                pixel[0] |= (c << maskOffsets[i]) & maskArray[i];
            }
        }
        switch (transferType) {
            case DataBuffer.BYTE_DATA: {
               byte bdata[];
               if (inData == null) {
                   bdata = new byte[1];
               } else {
                   bdata = (byte[])inData;
               }
               bdata[0] = (byte)(0xff&pixel[0]);
               return bdata;
            }
            case DataBuffer.SHORT_DATA:{
               short sdata[];
               if (inData == null) {
                   sdata = new short[1];
               } else {
                   sdata = (short[])inData;
               }
               sdata[0] = (short)(pixel[0]);
               return sdata;
            }
            case DataBuffer.INT_DATA:
               return pixel;
        }
        throw new NoSuchMethodError("This method has not been "+
                 "implemented for transferType " + transferType);

    }

    /**
     * Returns an array of all of the components given a pixel 
     * in the current color model. 
     * This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    final public int[] getComponents(int pixel, int[] components, int offset) {
        if (components == null) {
            components = new int[offset+numComponents];
        }

        for (int i=0; i < numComponents; i++) {
            components[offset+i] = (pixel & maskArray[i]) >>> maskOffsets[i];
        }

        return components;
    }
    
    /**
     * Returns an array of all of the components given a pixel array
     * in the current color model.   The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    final public int[] getComponents(Object obj, int[] components, int offset){
        int pixel=0;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])obj;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])obj;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])obj;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getComponents(pixel, components, offset);
    }
    
    /**
     * Creates a WritableRaster with the specified width and height, that 
     * has a channel layout compatible with this ColorModel.  
     * @see WritableRaster
     */
    final public WritableRaster createCompatibleWritableRaster (int w,
                                                                int h) {
        int[] bandmasks;
        if (supportsAlpha) {
            bandmasks = new int[4];
            bandmasks[3] = alpha_mask;
        }
        else {
            bandmasks = new int[3];
        }
        bandmasks[0] = red_mask;
        bandmasks[1] = green_mask;
        bandmasks[2] = blue_mask;
        
        if (pixel_bits > 16) {
	    return Raster.createPackedRaster(DataBuffer.INT_DATA,
                                             w,h,bandmasks,null); 
        }
        else if (pixel_bits > 8) {
	    return Raster.createPackedRaster(DataBuffer.SHORT_DATA,
                                             w,h,bandmasks,null); 
        }
        else {
	    return Raster.createPackedRaster(DataBuffer.BYTE_DATA,
                                             w,h,bandmasks,null); 
        }
    }

    /**
     * This method will pack the components into a pixel given an
     * array of discrete components in this color model.  This method
     * will throw an IllegalArgumentException if the components cannot
     * fit in a single integer pixel.
     */
    public int setComponents(int[] components, int offset) {
        int pixel = 0;
        for (int i=0; i < numComponents; i++) {
            pixel |= ((components[offset+i]<<maskOffsets[i])&maskArray[i]);
        }
        return pixel;
    }
    
    /**
     * This method will pack the components into a pixel given an
     * array of discrete components in this color model.
     * The pixel is
     * described by a discrete array of data elements of type 
     * transferType passed in as an object reference.
     * This array can be passed to the setPixelData methods of Raster.
     * objects.
     */
    public Object setComponents(int[] components, int offset, Object obj) {
        int pixel = 0;
        for (int i=0; i < numComponents; i++) {
            pixel |= ((components[offset+i]<<maskOffsets[i])&maskArray[i]);
        }
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               if (obj instanceof byte[]) {
                   byte bdata[] = (byte[])obj;
                   bdata[0] = (byte)(pixel&0xff);
                   return bdata;
               } else {
                   byte bdata[] = {(byte)(pixel&0xff)};
                   return bdata;
               }
            case DataBuffer.SHORT_DATA:
               if (obj instanceof short[]) {
                   short sdata[] = (short[])obj;
                   sdata[0] = (short)(pixel);
                   return sdata;
               } else {
                   short sdata[] = {(short)(pixel&0xff)};
                   return sdata;
               }
            case DataBuffer.INT_DATA:
               if (obj instanceof int[]) {
                   int idata[] = (int[])obj;
                   idata[0] = pixel;
                   return idata;
               } else {
                   int idata[] = {pixel};
                   return idata;
               }
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
    }

    /**
     * Forces the Raster data to match the state specified in the
     * isAlphaPremultiplied variable, assuming the data is currently
     * correctly described by this ColorModel.  It may multiply or divide
     * the color Raster data by alpha, or do nothing if the data is
     * in the correct state.  If the data needs to be coerced, this
     * method will also return an instance of this ColorModel with
     * the isAlphaPremultiplied flag set appropriately.
     */
    final public ColorModel coerceData (WritableRaster raster,
                                        boolean isAlphaPremultiplied)
    {
        if (!supportsAlpha ||
            this.isAlphaPremultiplied() == isAlphaPremultiplied) {
            return this;
        }
        
        int w = raster.getWidth();
        int h = raster.getHeight();
        int aIdx = numColorComponents;
        float normAlpha;
        int alphaScale = (1 << nBits[aIdx]) - 1;

        int rminX = raster.getMinX();
        int rY = raster.getMinY();
        int rX;
        int pixel[] = null;

        if (isAlphaPremultiplied) {
            // Must mean that we are currently not premultiplied so
            // multiply by alpha
            switch (transferType) {
                case DataBuffer.BYTE_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = raster.getPixel(rX, rY, pixel);
                            normAlpha = pixel[aIdx]/alphaScale;
                            if (normAlpha != 0.f) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] = (int)(pixel[c]*normAlpha);
                                }
                                raster.setPixel(rX, rY, pixel);
                            }
                        }
                    }
                }
                break;
                case DataBuffer.SHORT_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = raster.getPixel(rX, rY, pixel);
                            normAlpha = pixel[aIdx]/alphaScale;
                            if (normAlpha != 0.f) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] = (int)(pixel[c]*normAlpha);
                                }
                                raster.setPixel(rX, rY, pixel);
                            }
                        }
                    }
                } 
                break; 
                case DataBuffer.INT_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = raster.getPixel(rX, rY, pixel);
                            normAlpha = pixel[aIdx]/alphaScale;
                            if (normAlpha != 0.f) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] = (int)(pixel[c]*normAlpha);
                                }
                                raster.setPixel(rX, rY, pixel);
                            }
                        }
                    }
                }
                break; 
                default:
                    throw new NoSuchMethodError("This method has not been "+
                         "implemented for transferType " + transferType);
            }
        }
        else {
            // We are premultiplied and want to divide it out
            switch (transferType) {
                case DataBuffer.BYTE_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = raster.getPixel(rX, rY, pixel);
                            normAlpha = pixel[aIdx]/alphaScale;
                            if (normAlpha != 0) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] /= normAlpha;
                                }
                            }
                            raster.setPixel(rX, rY, pixel);
                        }
                    }
                }
                break;
                case DataBuffer.SHORT_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = raster.getPixel(rX, rY, pixel);
                            normAlpha = pixel[aIdx]/alphaScale;
                            if (normAlpha != 0) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] /= normAlpha;
                                }
                            }
                            raster.setPixel(rX, rY, pixel);
                        }
                    }
                }
                break;
                case DataBuffer.INT_DATA: {
                    for (int y = 0; y < h; y++, rY++) {
                        rX = rminX;
                        for (int x = 0; x < w; x++, rX++) {
                            pixel = raster.getPixel(rX, rY, pixel);
                            normAlpha = pixel[aIdx]/alphaScale;
                            if (normAlpha != 0) {
                                for (int c=0; c < aIdx; c++) {
                                    pixel[c] /= normAlpha;
                                }
                            }
                            raster.setPixel(rX, rY, pixel);
                        }
                    }
                }
                break;
                default:
                    throw new NoSuchMethodError("This method has not been "+
                         "implemented for transferType " + transferType);
            }
        }

        // Return a new color model
        return new DirectColorModel(colorSpace, pixel_bits, maskArray[0],
                                    maskArray[1], maskArray[2], maskArray[3],
                                    isAlphaPremultiplied);
                                    
    }

    /**
      * Returns true if raster is compatible with this ColorModel and
      * false if it is not.
      */
    public boolean isCompatibleRaster(Raster raster) {
        SampleModel sm = raster.getSampleModel();
        SinglePixelPackedSampleModel spsm;
        if (sm instanceof SinglePixelPackedSampleModel) {
            spsm = (SinglePixelPackedSampleModel) sm;
        }
        else {
            return false;
        }
        if (spsm.getNumBands() != getNumComponents()) {
            return false;
        }
	boolean flag = true;
	int[] bitMasks = spsm.getBitMasks();
	int[] bitSizes = spsm.getBitSizes();
        int[] bitOffsets = spsm.getBitOffsets();
	int totalBitSize = 0;
	
	for(int i=0; i<bitSizes.length; i++) {
            if (bitSizes[i]+bitOffsets[i] > totalBitSize) {
                totalBitSize = bitSizes[i]+bitOffsets[i];
            }
        }
	
	for (int i=0; i<numComponents; i++) {
	    if (bitMasks[i] != maskArray[i])
		flag = false;
	}
	    
        return ( (raster.getTransferType() == transferType) &&
		 (totalBitSize == pixel_bits ) && flag );
    }
    
    private void setFields() {
        // Set the private fields
        // REMIND: Get rid of these from the native code
        red_mask     = maskArray[0];
        red_offset   = maskOffsets[0];
        green_mask   = maskArray[1];
        green_offset = maskOffsets[1];
        blue_mask    = maskArray[2];
        blue_offset  = maskOffsets[2];
        if (nBits[0] < 8) {
            red_scale = (1 << nBits[0]) - 1;
        }
        if (nBits[1] < 8) {
            green_scale = (1 << nBits[1]) - 1;
        }
        if (nBits[2] < 8) {
            blue_scale = (1 << nBits[2]) - 1;
        }
        if (supportsAlpha) {
            alpha_mask   = maskArray[3];
            alpha_offset = maskOffsets[3];
            if (nBits[3] < 8) {
                alpha_scale = (1 << nBits[3]) - 1;
            }
        }
    }

    public String toString() {
        return new String("DirectColorModel: rmask="
                          +Integer.toHexString(red_mask)+" gmask="
                          +Integer.toHexString(green_mask)+" bmask="
                          +Integer.toHexString(blue_mask)+" amask="
                          +Integer.toHexString(alpha_mask));
    }
}

