/*
 * @(#)IndexColorModel.java	1.54 98/03/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image;

import java.awt.Transparency;
import java.awt.color.ColorSpace;

/**
 * A ColorModel class that represents pixel values which are
 * indices into a fixed colormap in the ColorModel's color space.
 * An optional transparent
 * pixel value can be supplied which indicates a completely transparent
 * pixel, regardless of any alpha value recorded for that pixel value.
 * This color model is similar to an X11 PseudoColor visual.  Note
 * that alpha values in IndexColorModels are never premultiplied.
 * <p>Many of the methods in this class are final.  The reason for
 * this is that the underlying native graphics code makes assumptions
 * about the layout and operation of this class and those assumptions
 * are reflected in the implementations of the methods here that are
 * marked final.  You can subclass this class for other reaons, but
 * you cannot override or modify the behaviour of those methods.
 *
 * @see ColorModel
 *
 * @version 10 Feb 1997
 */
public class IndexColorModel extends ColorModel {
    private int rgb[] = null;
    private int map_size = 0;
    private int transparent_index = -1;

    static private native void initIDs();
    static {
        ColorModel.loadLibraries();
        initIDs();
    }
    /**
     * Constructs an IndexColorModel from the given arrays of red,
     * green, and blue components.  Pixels described by this color
     * model will all have alpha components of 255 (fully opaque).
     * All of the arrays specifying the color components must have
     * at least the specified number of entries.  Uses the default
     * sRGB ColorSpace.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param r		The array of red color components.
     * @param g		The array of green color components.
     * @param b		The array of blue color components.
     */
    public IndexColorModel(int bits, int size,
			   byte r[], byte g[], byte b[]) {
	super(bits, ColorSpace.getInstance(ColorSpace.CS_sRGB),
              false, false, Transparency.OPAQUE);
        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
	setRGBs(size, r, g, b, null);
    }

    /**
     * Constructs an IndexColorModel from the given arrays of red,
     * green, and blue components.  Pixels described by this color
     * model will all have alpha components of 255 (fully opaque),
     * except for the indicated transparent pixel.  All of the arrays
     * specifying the color components must have at least the specified
     * number of entries.  Uses the default sRGB ColorSpace.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param r		The array of red color components.
     * @param g		The array of green color components.
     * @param b		The array of blue color components.
     * @param trans	The index of the transparent pixel.
     */
    public IndexColorModel(int bits, int size,
			   byte r[], byte g[], byte b[], int trans) {
	super(bits, ColorSpace.getInstance(ColorSpace.CS_sRGB),
              true, false, Transparency.BITMASK);
        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
	setRGBs(size, r, g, b, null);
        if (trans > -1) {
            transparency = Transparency.BITMASK;
            setTransparentPixel(trans);
        }
    }
 
    /**
     * Constructs an IndexColorModel from the given arrays of red,
     * green, blue and alpha components.  All of the arrays specifying
     * the color components must have at least the specified number
     * of entries.  Uses the default sRGB ColorSpace.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param r		The array of red color components.
     * @param g		The array of green color components.
     * @param b		The array of blue color components.
     * @param a		The array of alpha value components.
     */
    public IndexColorModel(int bits, int size,
			   byte r[], byte g[], byte b[], byte a[]) {
        super (bits, ColorSpace.getInstance(ColorSpace.CS_sRGB),
               true, false, Transparency.TRANSLUCENT);
        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
        setRGBs (size, r, g, b, a);
        setTransparentPixel (-1);
    }

    /**
     * Constructs an IndexColorModel from a single array of packed
     * red, green, blue and optional alpha components.  The array
     * must have enough values in it to fill all of the needed
     * component arrays of the specified size.  Uses the default
     * sRGB ColorSpace.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param cmap	The array of color components.
     * @param start	The starting offset of the first color component.
     * @param hasalpha	Indicates whether alpha values are contained in
     *			the cmap array.
     */
    public IndexColorModel(int bits, int size, byte cmap[], int start,
			   boolean hasalpha) {
	this(bits, size, cmap, start, hasalpha, -1);
        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
    }

    /**
     * Constructs an IndexColorModel from a single array of packed
     * red, green, blue and optional alpha components.  The specified
     * transparent index represents a pixel that will be considered
     * entirely transparent regardless of any alpha value specified
     * for it.  The array must have enough values in it to fill all
     * of the needed component arrays of the specified size.
     * Uses the default sRGB ColorSpace.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param cmap	The array of color components.
     * @param start	The starting offset of the first color component.
     * @param hasalpha	Indicates whether alpha values are contained in
     *			the cmap array.
     * @param trans	The index of the fully transparent pixel.
     */
    public IndexColorModel(int bits, int size, byte cmap[], int start,
			   boolean hasalpha, int trans) {
	// REMIND: This assumes the ordering: RGB[A]
	super(bits, ColorSpace.getInstance(ColorSpace.CS_sRGB),
              hasalpha ? true : (trans > 0 ? true : false),
              false,
              hasalpha ? Transparency.TRANSLUCENT
                       : (trans >= 0 ? Transparency.BITMASK
                                     : Transparency.OPAQUE) );
        
        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
	map_size = size;
	rgb = new int[Math.max(size, 256)];
	int j = start;
	int alpha = 0xff;
        transparency = OPAQUE;
	for (int i = 0; i < size; i++) {
	    rgb[i] = ((cmap[j++] & 0xff) << 16)
		| ((cmap[j++] & 0xff) << 8)
		| (cmap[j++] & 0xff);
	    if (hasalpha) {
		alpha = cmap[j++];
		if (alpha != 0xff && transparency != TRANSLUCENT) {
                    transparency =  (alpha == 0x0
                                     ? BITMASK
                                     : TRANSLUCENT);
		}
	    }
	    rgb[i] |= (alpha << 24);
	}

        setTransparentPixel(trans);
        if (transparent_index >= 0) {
            if (transparency == OPAQUE) {
                transparency = BITMASK;
            }
        }

        if (supportsAlpha) {
            nBits = new int[4];
            nBits[0] = nBits[1] = nBits[2] = nBits[3] = 8;
        }
        else {
            if (transparent_index > -1) {
                nBits = new int[4];
                nBits[3] = 1;
            }
            else {
                nBits = new int[3];
            }
            nBits[0] = nBits[1] = nBits[2] = 8;
        }
    }

    /**
     * Constructs an IndexColorModel from an array of integers
     * where each integer is comprised of
     * red, green, blue, and optional alpha components in the default
     * RGB color model packing.
     * The specified
     * transparent index represents a pixel that will be considered
     * entirely transparent regardless of any alpha value specified
     * for it.  The array must have enough values in it to fill all
     * of the needed component arrays of the specified size.
     * Uses the default sRGB ColorSpace.
     * @param bits	The number of bits each pixel occupies.
     * @param size	The size of the color component arrays.
     * @param cmap	The array of color components.
     * @param start	The starting offset of the first color component.
     * @param hasalpha	Indicates whether alpha values are contained in
     *			the cmap array.
     * @param trans	The index of the fully transparent pixel.
     * @param isAlphaPremultiplied Specifies how to interpret the color
     * data that are passed to this ColorModel.  If true, the data will
     * have been multiplied with alpha.
     */
    public IndexColorModel(int bits, int size,
                           int cmap[], int start,
			   boolean hasalpha, int trans,
                           boolean isAlphaPremultiplied) {
	// REMIND: This assumes the ordering: RGB[A]
	super(bits, ColorSpace.getInstance(ColorSpace.CS_sRGB),
              hasalpha ? true : (trans >= 0 ? true : false),
              isAlphaPremultiplied,
              hasalpha ? Transparency.TRANSLUCENT
                       : (trans >= 0 ? Transparency.BITMASK
                                    : Transparency.OPAQUE) );

        if (bits < 1 || bits > 32) {
            throw new IllegalArgumentException("Number of bits must be between"
                                               +" 1 and 32.");
        }
	map_size = size;
	rgb = new int[Math.max(size, 256)];
	int j = start;
	int alpha = 0xff000000;
        transparency = OPAQUE;
        if (!hasalpha) {
            // Need to make sure that the alpha is 0xff
            for (int i=0; i < size; i++) {
                rgb[i] = cmap[i] | 0xff000000;
            }
        }
        else {
            for (int i = 0; i < size; i++, j++) {
                rgb[i] = cmap[j];
                alpha = cmap[j] & 0xff000000;
                if (alpha != 0xff000000 && transparency != TRANSLUCENT) {
                    transparency =  (alpha == 0x0
                                     ? BITMASK
                                     : TRANSLUCENT);
                }
            }
	}

        setTransparentPixel(trans);
        if (transparent_index >= 0) {
            if (transparency == OPAQUE) {
                transparency = BITMASK;
            }
        }
    }

    private void setRGBs(int size, byte r[], byte g[], byte b[], byte a[]) {
	map_size = size;
	rgb = new int[Math.max(size, 256)];
	int alpha = 0xff;
        transparency = OPAQUE;
	for (int i = 0; i < size; i++) {
	    if (a != null) {
		alpha = (a[i] & 0xff);
		if (alpha != 0xff && transparency != TRANSLUCENT) {
                    transparency =  (alpha == 0x0
                                     ? BITMASK
                                     : TRANSLUCENT);
		}
	    }
	    rgb[i] = (alpha << 24)
		| ((r[i] & 0xff) << 16)
		| ((g[i] & 0xff) << 8)
		| (b[i] & 0xff);
	}
        nBits = new int[4];
        nBits[0] = nBits[1] = nBits[2] = nBits[3] = 8;
        maxBits = 8;
        
    }

    /**
     * Returns the transparency.  Returns either OPAQUE, BITMASK,
     * or TRANSLUCENT
     * @see Transparency#OPAQUE
     * @see Transparency#BITMASK
     * @see Transparency#TRANSLUCENT
     */
    public int getTransparency() {
        return transparency;
    }

    /**
     * Returns an array of the number of bits per component.  
     */
    public int[] getComponentSize() {
        if (nBits == null) {
            if (supportsAlpha) {
                nBits = new int[4];
                nBits[3] = 8;
            }
            else {
                nBits = new int[3];
            }
            nBits[0] = nBits[1] = nBits[2] = 8;
        }
        return nBits;
    }
    
    /**
     * Returns the size of the color component arrays in this
     * IndexColorModel.
     */
    final public int getMapSize() {
        return map_size;
    }

    /**
     * Returns the index of the transparent pixel in this IndexColorModel
     * or -1 if there is no transparent pixel.
     */
    final public int getTransparentPixel() {
	return transparent_index;
    }

    /**
     * Copies the array of red color components into the given array.  Only
     * the initial entries of the array as specified by getMapSize() are
     * written.  If the color space is not sRGB, then the entries are
     * converted to sRGB and the red components are extracted.
     */
    final public void getReds(byte r[]) {
        // REMIND: If not sRGB, need to convert
	for (int i = 0; i < map_size; i++) {
	    r[i] = (byte) (rgb[i] >> 16);
	}
    }

    /**
     * Copies the array of green color components into the given array.  Only
     * the initial entries of the array as specified by getMapSize() are
     * written.  If the color space is not sRGB, then the entries are
     * converted to sRGB and the green components are extracted.
     */
    final public void getGreens(byte g[]) {
	for (int i = 0; i < map_size; i++) {
	    g[i] = (byte) (rgb[i] >> 8);
	}
    }

    /**
     * Copies the array of blue color components into the given array.  Only
     * the initial entries of the array as specified by getMapSize() will
     * be written.  If the color space is not sRGB, then the entries are
     * converted to sRGB and the green components are extracted.
     */
    final public void getBlues(byte b[]) {
        for (int i = 0; i < map_size; i++) {
            b[i] = (byte) rgb[i];
        }
    }

    /**
     * Copies the array of alpha transparency values into the given array.
     * Only
     * the initial entries of the array as specified by getMapSize() will
     * be written.
     */
    final public void getAlphas(byte a[]) {
        for (int i = 0; i < map_size; i++) {
            a[i] = (byte) (rgb[i] >> 24);
        }
    }

    /**
     * Converts the arrays color and alpha components to the default
     * RGB ColorModel (if necessary) and copies the resulting array
     * of 32-bit ARGB values into the given array.  Only
     * the initial entries of the array as specified by getMapSize() will
     * be written.
     */
    final public void getRGBs(int rgb[]) {
        System.arraycopy(this.rgb, 0, rgb, 0, map_size);
    }

    private void setTransparentPixel(int trans) {
	if (trans >= map_size || trans < 0) {
	    trans = -1;
	} else {
	    rgb[trans] &= 0x00ffffff;
	}
	transparent_index = trans;
    }

    /**
     * Returns the red color component for the specified pixel in the
     * range 0-255 in the default RGB ColorSpace, sRGB.  
     * The returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    final public int getRed(int pixel) {
	return (rgb[pixel] >> 16) & 0xff;
    }

    /**
     * Returns the green color component for the specified pixel in the
     * range 0-255 in the default RGB ColorSpace, sRGB.  The
     * returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    final public int getGreen(int pixel) {
	return (rgb[pixel] >> 8) & 0xff;
    }

    /**
     * Returns the blue color component for the specified pixel in the
     * range 0-255 in the default RGB ColorSpace, sRGB.  The
     * returned value will be a
     * non pre-multiplied color (i.e. if the alpha is premultiplied,
     * this method will divide it out before it is returned).  If the
     * alpha value is 0, the color will be 0.
     */
    final public int getBlue(int pixel) {
	return rgb[pixel] & 0xff;
    }

    /**
     * Return the alpha transparency value for the specified pixel in the
     * range 0-255.
     */
    final public int getAlpha(int pixel) {
	return (rgb[pixel] >> 24) & 0xff;
    }

    /**
     * Returns the color of the pixel in the default RGB color model.
     * If the ColorSpace is not sRGB, it will be converted to sRGB.
     * @see ColorModel#getRGBdefault
     */
    final public int getRGB(int pixel) {
	return rgb[pixel];
    }

    /**
     * Returns a pixel array given a pixel in the default RGB color model.
     * The pixel is
     * described by a discrete array of data elements of type
     * transferType passed in as an object reference.
     * This pixel array can then be passed to the setPixelData method of
     * a WritableRaster object.  If the pixel variable is null, a new 
     * array will be allocated.  The pixel array will be returned.
     */
    public Object getPixelData(int rgb, Object pixel) {
        Object retObject = pixel;
        int red = (rgb>>16)&0xff;
        int green = (rgb>>8)&0xff;
        int blue  = rgb&0xff;
        int alpha = (rgb>>24)&0xff;
        int pix = 0;

        if (alpha == 0) {
            // Look for another transparent pixel
            if (transparent_index > -1) {
                pix = transparent_index;
            }
            else {
                int tmpalpha;
                // Search for one
                for (int i=0; i < map_size; i++) {
                    tmpalpha = (this.rgb[i]>>24)&0xff;
                    if (tmpalpha == 0) {
                        transparent_index = i;
                        pix = i;
                        break;
                    }
                }
            }
        } else {
            // a heuristic which says find the closest color,
            // after finding the closest alpha
            // if user wants different behavior, they can derive
            // a class and override this method
            // SLOW --- but accurate
            // REMIND - need a native implementation, and inverse color-map
            int tmpalpha;
            int tmpred;
            int tmpblue;
            int tmpgreen;
            int alphaScale = 3 * (255 * 255);       //alpha is most important
            int smallestError = 255 * alphaScale;   // largest possible
            int smallestIndex = 0;
            int currentError;
            for (int i=0; i < map_size; i++) {
                tmpalpha = (this.rgb[i]>>24)&0xff;
                tmpred = (this.rgb[i]>>16)&0xff;
                tmpgreen = (this.rgb[i]>>8)&0xff;
                tmpblue = (this.rgb[i])&0xff;
                currentError = (alpha-tmpalpha) * (alpha-tmpalpha);
                currentError *= alphaScale;
                currentError += (red - tmpred) * (red - tmpred);
                currentError += (green - tmpgreen) * (green - tmpgreen);
                currentError += (blue - tmpblue) * (blue - tmpblue);
                if (currentError < smallestError) {
                    smallestIndex = i;
                }
            }
            pix = smallestIndex;
        }

        if (red == green && green == blue) {
            // Grayscale
        }
        switch (transferType) {
        case DataBuffer.INT_DATA:
            if (pixel == null) {
                int[] obj = new int[1];
                obj[0] = pix;
                pixel = obj;
                retObject = obj;
            }
            else {
                int[] obj = (int[]) pixel;
                obj[0] = pix;
                retObject = obj;
            }
            break;
        case DataBuffer.BYTE_DATA:
            if (pixel == null) {
                byte[] obj = new byte[1];
                obj[0] = (byte) pix;
                pixel =  obj;
                retObject = obj;
            }
            else {
                byte[] obj = (byte[]) pixel;
                obj[0] = (byte) pix;
                retObject = obj;
            }
            break;
        case DataBuffer.SHORT_DATA:
            if (pixel == null) {
                short[] obj = new short[1];
                obj[0] = (short) pix;
                pixel = obj;
                retObject = obj;
            }
            else {
                short[] obj = (short[]) pixel;
                obj[0] = (short) pix;
                retObject = obj;
            }
            break;
        default:
            throw new IllegalArgumentException("This method has not been "+
                             "implemented for transferType " + transferType);
        }
        return retObject;
    }

    /**
     * Returns an array of all of the components given a pixel 
     * in the current color model.  
     * This can be used to retrieve the
     * components for a specific pixel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    public int[] getComponents(int pixel, int[] components, int offset) {
        if (components == null) {
            components = new int[offset+4];
        }

        // REMIND: Needs to change if different color space
        components[offset+0] = getRed(pixel);
        components[offset+1] = getGreen(pixel);
        components[offset+2] = getBlue(pixel);
        if ((components.length-offset) > 3) {
            components[offset+3] = getAlpha(pixel);
        }
        
        return components;
    }
    
    /**
     * Returns an array of all of the components given a pixel array
     * in the current color model.   The pixel is
     * described by a discrete array of data elements of type
     * transferType passed in as an object reference.
     * This can be used to retrieve the
     * components for a specific pixel in a PackedColorModel.  If the
     * components array is null, a new array will be allocated.  The
     * components array will be returned.
     */
    public int[] getComponents(Object inPixel, int[] components, int offset) {
        int pixel;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inPixel;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inPixel;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inPixel;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return getComponents(pixel, components, offset);
    }
    
    /**
     * Returns a pixel, given a discrete components (unpacked color/alpha
     * values) in this color model.
     */
    public int setComponents(int[] components, int offset) {
        int rgb = (components[offset+0]<<16)
            | (components[offset+1]<<8) | (components[offset+2]);
        if (supportsAlpha) {
            rgb |= (components[offset+3]<<24);
        }
        else {
            rgb &= 0xff000000;
        }
        Object inData = getPixelData(rgb, null);
        int pixel;
        switch (transferType) {
            case DataBuffer.BYTE_DATA:
               byte bdata[] = (byte[])inData;
               pixel = bdata[0] & 0xff;
            break;
            case DataBuffer.SHORT_DATA:
               short sdata[] = (short[])inData;
               pixel = sdata[0];
            break;
            case DataBuffer.INT_DATA:
               int idata[] = (int[])inData;
               pixel = idata[0];
            break;
            default:
               throw new NoSuchMethodError("This method has not been "+
                   "implemented for transferType " + transferType);
        }
        return pixel;
    }
    
    /**
     * Returns a pixel array, given a discrete components (unpacked
     * color/alpha values) in this color model.   The pixel is
     * described by a discrete array of data elements of type
     * transferType passed in as an object reference.
     * This array can be passed to the setPixelData methods of Raster.
     * objects.
     */
    public Object setComponents(int[] components, int offset, Object pixel) {
        int rgb = (components[offset+0]<<16) | (components[offset+1]<<8)
            | (components[offset+2]);
        if (supportsAlpha) {
            rgb |= (components[offset+3]<<24);
        }
        else {
            rgb &= 0xff000000;
        }
        return getPixelData(rgb, pixel);
    }

    /**
     * Creates a WritableRaster with the specified width and height, that 
     * has a channel layout compatible with this ColorModel.  
     * @see WritableRaster
     */
    public WritableRaster createCompatibleWritableRaster (int w, int h) {
        WritableRaster raster;

        if (pixel_bits == 1) {
            // TYPE_BINARY
            raster = Raster.createPackedRaster(DataBuffer.BYTE_DATA,
                                               w, h, 1, 1, null);
        }
        else if (pixel_bits <= 8) {
            raster = Raster.createComponentRaster(DataBuffer.BYTE_DATA,
                                                  w,h,1,null);
        }
        else if (pixel_bits <= 16) {
            raster = Raster.createComponentRaster(DataBuffer.SHORT_DATA,
                                                  w,h,1,null);
        }
        else {
            raster = Raster.createComponentRaster(DataBuffer.INT_DATA,
                                                  w,h,1,null);
        }
        return raster;
    }

    /**
     * Forces the Raster data to match the state specified in the
     * isAlphaPremultiplied variable, assuming the data is currently
     * correctly described by this ColorModel.  It may multiply or divide
     * the color Raster data by alpha, or do nothing if the data is
     * in the correct state.  If the data needs to be coerced, this
     * method will also return an instance of this ColorModel with
     * the isAlphaPremultiplied flag set appropriately.
     */
    public ColorModel coerceData (WritableRaster raster, 
                                  boolean isAlphaPremultiplied) {
        if ((supportsAlpha == false) ||
            (this.isAlphaPremultiplied == isAlphaPremultiplied))
        {
            // Nothing to do
            return this;
        }

        int[] newlut = new int[map_size];
        int red, grn, blu, alp;
        if (this.isAlphaPremultiplied) {
            for (int i=0; i < map_size; i++) {
                alp = (rgb[i]>>24)&0xff;
                red = (rgb[i]>>16)&0xff;
                grn = (rgb[i]>>8) &0xff;
                blu = (rgb[i])    &0xff;
                if (alp == 0) {
                    red = grn = blu = 0;
                }
                else if (alp < 255) {
                    if (alp > 128) {
                        alp++;
                    }
                    red = (red*alp)>>8;
                    grn = (grn*alp)>>8;
                    blu = (blu*alp)>>8;
                }
                newlut[i] = rgb[i]&0xff000000 |
                    ((red&0xff) << 16) | ((grn&0xff) << 8) | (blu&0xff);
            }
        }
        else {
            // Want it not premultiplied
            for (int i=0; i < map_size; i++) {
                alp = (rgb[i]>>24)&0xff;
                red = (rgb[i]>>16)&0xff;
                grn = (rgb[i]>>8) &0xff;
                blu = (rgb[i])    &0xff;
                if (alp > 0 && alp < 255) {
                    if (alp > 128) {
                        alp++;
                    }
                    red = (red<<8)/alp;
                    grn = (grn<<8)/alp;
                    blu = (blu<<8)/alp;
                }
                newlut[i] = rgb[i]&0xff000000 |
                    ((red&0xff) << 16) | ((grn&0xff) << 8) | (blu&0xff);
            }

        }
        return new IndexColorModel(pixel_bits, map_size, newlut, 0,
                                   supportsAlpha, transparent_index,
                                   isAlphaPremultiplied);
    }

    /**
      * Returns true if raster is compatible with this ColorModel and
      * false if it is not.
      */
    public boolean isCompatibleRaster(Raster raster) {

	int size = raster.getSampleModel().getSampleSize(0);
	
        return ((raster.getTransferType() == transferType) &&
		(raster.getNumBands() == 1) && ((1 << size) >= map_size));
    }
    
    /**
     * Returns a new BufferedImage of TYPE_INT_ARGB or TYPE_INT_RGB that
     * matches the lookup tables.
     */
    public BufferedImage convertToIntDiscrete(Raster raster, 
                                              boolean forceARGB) {
        ColorModel cm;

        if (forceARGB || transparency == TRANSLUCENT) {
            cm = ColorModel.getRGBdefault();
        }
        else if (transparency == BITMASK) {
            cm = new DirectColorModel(25, 0xff0000, 0x00ff00, 0x0000ff,
                                      0x1000000);
        }
        else {
            cm = new DirectColorModel(24, 0xff0000, 0x00ff00, 0x0000ff);
        }

        int w = raster.getWidth();
        int h = raster.getHeight();
        WritableRaster discreteRaster = 
                  cm.createCompatibleWritableRaster(w, h);
        Object obj = null;
        int[] data = null;

        // REMIND:  Don't need this if we decide not to support
        // arbitrary color spaces
        if (colorSpace != ColorSpace.getInstance(ColorSpace.CS_sRGB)) {
            ColorConvertOp ccop = new ColorConvertOp();
            BufferedImage src = new BufferedImage(this, (WritableRaster)raster,
                                                  isAlphaPremultiplied);
            BufferedImage dst = new BufferedImage(cm, discreteRaster,
                                                  false);

            return ccop.filter(src, dst);
        }
            
        int rX = raster.getMinX();
        int rY = raster.getMinY();

        for (int y=0; y < h; y++, rY++) {
            obj = raster.getPixelData(rX, rY, w, 1, obj);
            if (obj instanceof int[]) {
                data = (int[])obj;
            } else {
                data = DataBuffer.toIntArray(obj);
            }
            for (int x=0; x < w; x++) {
                data[x] = rgb[data[x]];
            }
            discreteRaster.setPixelData(0, y, w, 1, data);
        }
        
        return new BufferedImage(cm, discreteRaster, false);
    }

}

