/*
 * @(#)JPEGImageDecoder.java	1.5 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/
package java.awt.image.codec;

    /**
     * Class JPEGImageDecoder 
     * 
     * JPEGImageDecoder decompresses an JPEG InputStreams into a Raster
     * or a BufferedImage depending upon the method invoked. The
     * JPEGImageDecoder can also act as an ImageProducer sending each
     * scanline to image consumers.  Decoding the JPEG input stream is
     * controlled by the parameters in theJPEGParam
     * object.  If no JPEGParam object is specified when JPEGImageDecoder 
     * is constructed, the constructor creates a JPEGParam object that has 
     * decompression options set to the defaults for the IJG implementation.
     * 
     * The JPEGParam object is updated with information from the file header
     * during decompression. If the input stream contains tables only
     * information (no image data), the JPEGParam object will be updated and
     * NULL returned for the output image. If the input stream contains only
     * image data, the parameters and tables in the current JPEGParam object
     * will be used to decode in decoding the JPEG stream. If no tables are 
     * set in the JPEGParam  object, an exception will be thrown.
     *
     * ColorSpace comments:  First off JPEG by specification is color blind!
     * That said, the IJG jpeg library does perform some color space conversion 
     * in the name of better compression ratios.  If the color space of the data
     * that is passed to the decoder has been converted the decoder will undo
     * the conversion if it knows how.  Some updates to the standard color space
     * designations have been made to allow this decoder to be utilized for
     * decoding FlashPix Images.  I the case where these color spaces are 
     * encountered no color space conversion is performed.  The data is simply
     * decoded by the library.  See the JPEGParam description for more details
     * on additional color space designations.
     *
     * SampleModel/DataModel comments:  First off the best performance from
     * a strictly decompression point of view can be had by using methods that
     * request the native library to return data in a ComponentSampleModel
     * that contains DataBuffer.BYTE_DATA.   Additionally, a 
     * BufferedImage can be returned from the decoder.  If possible this 
     * BufferedImage will contain an array of int's that contain packed 
     * ARGB data.  If the data cannot be coereced into this form an exception
     * will be thrown.  This particular data organization
     * ( BufferedImage.TYPE_INT_ARGB ) will facilitate easy display of the 
     * decoded images within the java2d framework.
     * 
     * Final Comments:  This decoder will process interchange, abbreviated and 
     * progressive jpeg streams.  However, progressive jpeg streams are treated
     * as interchange streams.  They return once with the entire image in the 
     * image buffer.  
     */

import java.io.InputStream;
import java.io.IOException;
import java.util.Vector;
import java.awt.Point;
import java.awt.image.*;
import java.awt.color.*;
import java.util.Enumeration;

/**
 * This class describes a JPEG image decoder.  This decoder takes an InputStream
 * that contains JPEG encoded image data.  The JPEGImageDecoder will decode the
 * JPEG image data according to the parameters set in a JPEGParam object.  The 
 * resulting image data is returned in either a Raster or a BufferedImage. 
 * @see JPEGParam
 * @see Raster
 * @see BufferedImage
 * @version 4 December 1997
 */
 
public class JPEGImageDecoder implements ImageProducer {

    private static final Class InputStreamClass = InputStream.class;
    /** 
     * Private Stuff 
     *
     */
    private JPEGParam 		jpegParam;
    private InputStream         input = null;
    private Raster		aRaster = null;
    private BufferedImage       aBufferedImage;
    private boolean             generateBufferedImage = false;
    private boolean		packData = false;
    private boolean             feedImgConsumer = false;
    private boolean             jpegHeaderProcessed = false;
    private boolean 		use_tables = true;
    private ColorModel		cm;
    private int			NbOfComponents_out  = 0;
    private int 		width = 0;
    private int 		height = 0;
    private Vector 		ics;


    /**
     * Note: The native methods described below have the potential to 
     * throw the following Exceptons:
     *      IllegalArgumentException
     *      NullPointerException
     *      OutOfMemoryError
     *      ArrayIndexOutOfBoundsException
     *      ImageFormatException
     *
     * It is expected that users of this class will handle these 
     * appropriatly
     */

    // native method that will initialize the method id's for all calls back
    // to java.
    private native void initDecoder( Class inputStrClass);


    private synchronized native void readJPEGStream(InputStream is,
						    JPEGParam params,
						    boolean use_tables, 
						    boolean packData,
						    boolean feedImgConsumer )
	throws IOException, ImageFormatException; 
		
    // load the jpeg lib when created.
    static { 
        try {
	    java.security.AccessController.beginPrivileged();
	    System.loadLibrary("jpeg" );
	} 
	finally {
	    java.security.AccessController.endPrivileged();
	}
    }

    /**
     * Constructor for those wishing to push pixels one line at a time
     * By design/assumption this will push packed int lines of pixels
     */
    public JPEGImageDecoder(InputStream in, ImageConsumer ic ) {
 	ics = new Vector();
	initDecoder( InputStreamClass );
	input = in;
	jpegParam = new JPEGParam();
	packData = true;
	generateBufferedImage = true;
	feedImgConsumer = true;
	initDecoder( InputStreamClass );
	addConsumer(ic);
    }

    /**
     * Constructs a JPEGImageDecoder that will decode the JPEG data 
     * found in the InputStream.  The default decompression parameters will 
     * be set.
     * @param in an InputStream containing JPEG encoded data
     * @see InputStream
     */
    public JPEGImageDecoder(InputStream in) {
	input = in;
	jpegParam = new JPEGParam();
	initDecoder( InputStreamClass );
    }

    /**
     * Constructs a JPEGImageDecoder that will decode the jpeg stream 
     * found in the input stream.  The decompression will be performed according
     * to the values found in the JPEGParam passed.
     * @param in an InputStream containing JPEG encoded data
     * @param jp a JPEGParam object containing decompression information.
     * @see InputStream
     * @see JPEGParam
     */
    public JPEGImageDecoder(InputStream in , JPEGParam jp) {
	input = in;
	jpegParam = jp;
	initDecoder( InputStreamClass );
    }	

    /**
     * Returns the JPEGParam object containing the JPEG tables and 
     * parameters
     */
    public JPEGParam getJPEGParam() {
        return jpegParam;
    }

    /**
     * Sets the JPEGParam object used to determine the features of the
     * decompression performed on the JPEG encoded data.
     * @param jp JPEGParam object
     */
    public void setJPEGParam(JPEGParam jp) {
        jpegParam = jp;
    }
	
    /** Creates the output image with properties read from the JPEG
     * header and initializes the JPEGParam object, but does not
     * actually decompress the image data. Throws an exception if
     * the input stream contains only image data. 
     */

    private void readTables() {
	try{
            //jpegParam.setwriteImage( false );
            //jpegParam.setwriteTables( true );
            readJPEGStream(input, jpegParam, false, packData,
                feedImgConsumer);
	}
	catch( ImageFormatException ife ){ ife.printStackTrace();}
	catch(IOException ioe){ioe.printStackTrace();}
	jpegHeaderProcessed = true;
    }

    /**
     * decodes the JPEG stream that was passed as part of construction.  The 
     * JPEG decompression will be performed according to the current settings 
     * of the JPEGParam object.
     * @return Raster the resulting image will be returned in a byte component
     *         Raster of data.  Colorspace and other pertinent information can 
     *         be obtained from the JPEGParam object.
     * @exception ImageFormatException if irregularities in the JPEG stream or 
     *         an unknown condition was encountered an ImageFormatException is
     *         thrown.
     */
    public synchronized  Raster decodeRaster() throws ImageFormatException { 
	try {
	    readJPEGStream(input, jpegParam, false, packData, feedImgConsumer);
        }
	catch(IOException e){ 
	    System.out.println("Can't open input Stream" + e);
	    e.printStackTrace();
	}
	catch(ImageFormatException ife ) {
            throw(ife);
        }
	return aRaster;
    }

    /**
     * Decodes the JPEG stream that is passed in as a parameter.  This method is
     * usually invoked when processing abbreviated JPEG data streams.  The 
     * JPEG decompression will be performed according to the current settings 
     * of the JPEGParam object.
     * @return Raster the resulting image will be returned in a byte component
     *         Raster of data.  Colorspace and other pertinent information can 
     *         be obtained from the JPEGParam object.
     * @exception ImageFormatException if irregularities in the JPEG stream or 
     *         an unknown condition was encountered an ImageFormatException is
     *         thrown.
     */
    public synchronized  Raster decodeRaster(InputStream isIn)
	throws ImageFormatException { 
	try {
	    readJPEGStream(isIn, jpegParam, true, packData, feedImgConsumer );
	}
        catch(IOException e) {
	    //System.out.println("Can't open input Stream" + e);
	}
	catch( ImageFormatException ife ){
	    //System.out.println(
	  //"The JPEG Lib received a JPEG stream that it couldn't understand.");
	    //System.out.println(ife);
	    throw(ife);
	}
	return aRaster;
    }

    /**
     * Decodes the JPEG data stream that was passed in the constructor.  The
     * result of decoding this InputStream is a BufferedImage that has a color
     * model of argb.
     * @return BufferedImage - contains the image data in an INT_ARGB format.
     * @exception ImageFormatException -  if irregularities in the JPEG stream
     *     or an unknown condition was encountered an ImageFormatException is
     *     thrown.
     */
    public synchronized  BufferedImage decodeBufferedImage()
	throws ImageFormatException { 
	generateBufferedImage = true;
	packData = true;
	try {
	    readJPEGStream(input, jpegParam, false, packData, feedImgConsumer);
	}
	catch( ImageFormatException ife ){
	    //System.out.println(
	  //"The JPEG Lib received a JPEG stream that it couldn't understand.");
	    //System.out.println(ife);
	    throw(ife);
	}
	catch(IOException e){
	    //System.out.println("Can't open input Stream" + e);
	}			
	return aBufferedImage;
    }

    /**
     * Decodes the JPEG data stream thatis passed in as a parameter.  The result
     * of decoding this InputStream is a BufferedImage that has a color model
     * of argb.
     * @return BufferedImage - contains the image data in an INT_ARGB format.
     * @exception ImageFormatException -  if irregularities in the JPEG stream
     *      or an unknown condition was encountered an ImageFormatException is
     *      thrown.
     */
    public synchronized  BufferedImage decodeBufferedImage(InputStream inStr)
	throws ImageFormatException { 
	generateBufferedImage = true;
	packData = true;
	try {
	    readJPEGStream(inStr, jpegParam, true, generateBufferedImage,
                           feedImgConsumer);
	}
	catch( ImageFormatException ife) {
	    //System.out.println(
	  //"The JPEG Lib received a JPEG stream that it couldn't understand.");
	    //System.out.println(ife);
	    throw(ife);
	}
	catch(IOException e){
	    //System.out.println("Can't open input Stream" + e);
	}			
	return aBufferedImage;
    }

    /**
     * allocateDataBuffer - This is a private method that is generally
     * intended to be called from native code after the JPEG header has been
     * read.  This provides the information necessary to create the
     * DataBufferByte object that will actually contain the decoded image
     * information.
     * @param imgWidth
     * @param imgHeight
     * @param imgComponents
     * @return byte[] - an array in which the decoded image data will be stored.
     */
    private byte[] allocateDataBuffer(int imgWidth ,int imgHeight,
                                      int imgComponents) {
	// this is the first time we will have height & width info.  Lets tell
        // our friends
	if (feedImgConsumer == true )
	    updateConsumers();

	if(generateBufferedImage == true ) {
	    if(jpegParam.getImageColorID() ==
               JPEGParam.COLOR_ID_GRAYSCALE ) {
	        aBufferedImage = new BufferedImage(imgWidth, imgHeight, 
				         BufferedImage.TYPE_BYTE_GRAY);
		aRaster = aBufferedImage.getRaster();
	    }
	    else {
		// if a buffered image is desired and the decoded colorspace is
		// not grayscale then asking for an array of bytes is the wrong
                // thing to do.
		return null;
	    }
	}
	else {
	    aRaster = Raster.createComponentRaster(DataBuffer.BYTE_DATA,
						   imgWidth, imgHeight,
						   imgComponents,
						   new Point(0,0));
        }
	if(generateBufferedImage == true )
	    cm = aBufferedImage.getColorModel();
	return  ((DataBufferByte)aRaster.getDataBuffer()).getData();
    }

    /**
     * allocateDataBuffer - allocates a BufferedImage of type
     * BufferedImage.TYPE_INT_ARGB and returns a reference to the int
     * array that will eventually hold the decoded image data.
     * @param imgWidth
     * @param imgHeight
     * @return int[] - an array in which the decoded image data will be stored.
     */
    private int[] allocateDataBuffer(int imgWidth ,int imgHeight){
	// this is the first time we will have height & width info.  Lets tell
        // our friends
	if (feedImgConsumer == true )
	    updateConsumers();

	int decodedID = jpegParam.getImageColorID();
	int numComponents = jpegParam.getNumComponents(decodedID);
	switch( numComponents ) {

	case 1: { // grayscale image
	    if(generateBufferedImage == true ) {
		// can't create a buffered image with grayscale int buffers
		// the allocateDataBuffer that returns a byte[] should have been
		// called
		return null;
	    }
	    else {
		aRaster = Raster.createComponentRaster(DataBuffer.INT_DATA,
						       imgWidth,imgHeight,
                                                       numComponents,
                                                       new Point(0,0));
	    }
	    break;
	}

	case 3: { // 	
	    if(generateBufferedImage == true ) {
		if(packData == true ) {
		    aBufferedImage = new BufferedImage(imgWidth, imgHeight, 
			   		     BufferedImage.TYPE_INT_ARGB);
                }
		else {
		    aBufferedImage = new BufferedImage(imgWidth, imgHeight,
                                             BufferedImage.TYPE_INT_RGB);
                }
		aRaster = aBufferedImage.getRaster();
	    }
	    else {
		aRaster = Raster.createComponentRaster(DataBuffer.INT_DATA,
                                                       imgWidth, imgHeight,
                                                       numComponents,
                                                       new Point(0,0));
	    }
	    break;
	}

	case 4:
	    if(generateBufferedImage == true) {
				//should never reach here because
		return null;	// we already checked hasKnowColorModel();
            }
	    if(packData == true ) {
		aRaster = Raster.createComponentRaster(DataBuffer.INT_DATA,
                                                       imgWidth, imgHeight,
                                                       1 /* ??? */,
                                                       new Point(0,0));
            }
	    else {
		aRaster = Raster.createComponentRaster(DataBuffer.INT_DATA,
                                                       imgWidth, imgHeight,
                                                       numComponents,
                                                       new Point(0,0));
            }
	}  // end switch

	if(generateBufferedImage == true) {
	    cm = aBufferedImage.getColorModel();
        }
	return ((DataBufferInt)aRaster.getDataBuffer()).getData();
    }


    ///////////////////////////////////////////////////////////////////////////
    //
    //			Image Producer support
    //
    ///////////////////////////////////////////////////////////////////////////
    /**
     * sendPixels - intended to be called by native code after each scanline
     * of jpeg image data has been decoded.   
     * @param int[] - the array containing the pixels returned by the
     *                decompressor
     * @param int - the row number corresponding to this row of information
     *              in the image.
     * @return boolean indicating whether the transfer of pixels was successful
     */	
    private boolean sendPixelsInt( int[] pixels, int rowNum ) {
	Enumeration icList;
	ImageConsumer ic;
	int width = jpegParam.getWidth();
	int height = jpegParam.getHeight();
	// Set up the ImageConsumers
	icList = ics.elements();
	while(icList.hasMoreElements()) {
	    ic = (ImageConsumer)icList.nextElement();
	    ic.setPixels(0, rowNum, width , 1, cm, pixels, 0, width );
	    if(rowNum == height) {
		ic.setHints(ImageConsumer.TOPDOWNLEFTRIGHT |
			    ImageConsumer.COMPLETESCANLINES |
			    ImageConsumer.SINGLEFRAME |
			    ImageConsumer.STATICIMAGEDONE );
            }
	}
	return true;
    }

    /**
     * sendPixels - intended to be called by native code after each scanline
     * of jpeg image data has been decoded.   
     * @param byte[] - the array containing the pixels returned by the
     *                 decompressor
     * @param int - the row number corresponding to this row of information
     *              in the image.
     * @return boolean indicating whether the transfer of pixels was successful
     */	
    private boolean sendPixelsByte( byte[] pixels, int rowNum){
	Enumeration icList;
	ImageConsumer ic;
	// Set up the ImageConsumers
	icList = ics.elements();
	while(icList.hasMoreElements()) {
	    ic = (ImageConsumer)icList.nextElement();
	    ic.setPixels(0, rowNum, width , 1, cm, pixels, 0,
                         (width * NbOfComponents_out) );
	    if(rowNum == height) {
		ic.setHints(ImageConsumer.TOPDOWNLEFTRIGHT |
			    ImageConsumer.COMPLETESCANLINES |
			    ImageConsumer.SINGLEFRAME |
			    ImageConsumer.STATICIMAGEDONE );
            }
	}
	return true;

    }

    /**
     * Adds an ImageConsumer to the list of consumers interested in
     * data for this image.
     * @see ImageConsumer
     */
    public synchronized void addConsumer(ImageConsumer ic) {
    	if( feedImgConsumer == false) {
    	    ics = new Vector();
    	    feedImgConsumer = true;
    	}
	if (ics.contains(ic)) {
	    return;
	}
	ics.addElement(ic);
    }

    /**
     * Determine if an ImageConsumer is on the list of consumers currently
     * interested in data for this image.
     * @return true if the ImageConsumer is on the list; false otherwise
     * @see ImageConsumer
     */
    public synchronized boolean isConsumer(ImageConsumer ic) {
	return ics.contains(ic);
    }

    /**
     * Remove an ImageConsumer from the list of consumers interested in
     * data for this image.
     * @see ImageConsumer
     */
    public synchronized void removeConsumer(ImageConsumer ic) {
	ics.removeElement(ic);
    }

    /**
     * Adds an ImageConsumer to the list of consumers interested in
     * data for this image, and immediately start delivery of the
     * image data through the ImageConsumer interface.
     * @see ImageConsumer
     */
    public synchronized void startProduction(ImageConsumer ic) {
	if( input != null ){
	    addConsumer(ic);
	    updateConsumers();
	    decodeBufferedImage();
	}
    }

    /**
     * Requests that a given ImageConsumer have the image data delivered
     * one more time in top-down, left-right order.
     * @see ImageConsumer
     */
    public void requestTopDownLeftRightResend(ImageConsumer ic) {
	// So far, all pixels are already sent in TDLR order
    }

    /**
     * Called from native code when the decoder is asked to act as an
     * ImageProducer.  This is called at the first opportunity that the
     * jpeg lib has knowledge of these parameters.
     */
    private void updateConsumers(){
	Enumeration icList;
	ImageConsumer consumer;
	int imageID = jpegParam.getImageColorID();
	width = jpegParam.getWidth();
	height = jpegParam.getHeight();
	if(( width > 0) & (height > 0) ){
	    NbOfComponents_out = jpegParam.getNumComponents(imageID);

	    // setup our color model
	    switch(imageID) {

	    case JPEGParam.COLOR_ID_GRAYSCALE: {
		byte g[] = new byte[256];
		for (int i = 0; i < 256; i++) {
		    g[i] = (byte) i;
		}
		cm = new IndexColorModel(8, 256, g, g, g);
		break;
	    }

	    case JPEGParam.COLOR_ID_RGB: {
		if(packData)
		    cm =  ColorModel.getRGBdefault();
		else
		    cm = new DirectColorModel(24, 0xff0000, 0xff00, 0xff);
		break;
	    }

	    default:
		throw(new ImageFormatException(
                      "Invalid colorspace for generating BufferedImage"));
				
            }

	    icList = ics.elements();

	    while(icList.hasMoreElements()) {
	        consumer = (ImageConsumer)icList.nextElement();
	        consumer.setDimensions(jpegParam.getWidth(),
                                       jpegParam.getHeight());
	        consumer.setHints(ImageConsumer.TOPDOWNLEFTRIGHT |
			          ImageConsumer.COMPLETESCANLINES |
			          ImageConsumer.SINGLEFRAME);
	    }
	}
    }

} // end class JPEGImageDecoder

