/*
 * @(#)JPEGImageEncoder.java	1.4 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image.codec; 

/**
 * Class JPEGImageEncoder 
 * 
 * JPEGImageEncoder compresses images into JPEG data streams and writes
 * the JPEG stream to an OutputStream.  Image data that is to be encoded
 * can be passed in as a Raster of image data or as a BufferedImage.
 * Encoding or the image data into the output JPEG stream is controlled
 * by the parameters setting found in the JPEGParam object.  If no JPEGParam 
 * object is specified when JPEGImageEncoder is constructed, the constructor 
 * creates a JPEGParam object that has default compression settings according
 * to the defaults for the IJG implementation.
 * 
 *
 * ColorSpace comments:  First off JPEG by specification is color blind!
 * That said, the IJG jpeg library does perform some color space conversion 
 * in the name of better compression ratios.  There is no explicit mechanism
 * in the JPEGParam interface for controlling the Encoded ColorSpace of the 
 * data when it is written to the JPEG data stream.  Generally, users should 
 * simply tell the truth when setting the "DecodedColorspace" parameter of 
 * the JPEGParam object.  If an approriate colorspace setting is not already 
 * defined it is recommended that colorspace unknown is used.  Some updates 
 * to the standard color space designations have been made to allow this 
 * decoder to be utilized for decoding FlashPix Images.  I the case where 
 * these color spaces are encountered no color space conversion is performed.  
 * The data is simply decoded by the library.  See the JPEGParam description 
 * for more details on additional color space designations.
 *
 * SampleModel/DataModel comments:  First off the best performance from
 * a strictly compression point of view can be had by using methods that
 * contain image data in a in a ComponentSampleModels.  Packed data will 
 * be unpacked into byte components before being passed to the JPEG compressor.
 * Finally, as a special case, data that is stored in a BufferedImage in a 
 * packed int ARGB data model will have the Alpha channel removed when the 
 * unpacking is being performed.  This will result in jpeg image streams that
 * correspond to the JFIF standard.  If it is desired that BufferedImages 
 * data be compressed with the alhpa channel set the decoded colorspace 
 * parameter of the JPEGParam object to something other than IJG_JCS_RGB.
 * Note, it is recommended that the alhpa data be premultiplied into the 
 * RGB channels before encoding is performed.  Note, removal of the alpha
 * channel only applies to images passed in BufferedImages.
 * 
 * Final Comments:  This encoder will process interchange, and abbreviated 
 * jpeg streams.  
 */

import java.io.*;
import java.awt.image.*;

/**
 * JPEGImageEncoder encodes buffers of image data into JPEG data
 * streams.  Essentially users of this class are required to provide image
 * data in a Raster object or a BufferedImage, set the necessary parameters
 * in the JPEGParams object and successfully open the 
 * <code>OutputStream</code> that is the destination of the encoded jpeg 
 * stream.  
 *
 * The JPEGImageEncoder class will encode Rasters of image data into
 * interchange, and abbreviated JPEG data streams that are written to the
 * OutputStream provided to the encoder.
 *
 * 
 * @see		JPEGParam
 * @see		Raster
 * @see		BufferedImage
 * @see		OutputStream
 */

public class JPEGImageEncoder {
	/** Object that contains the details of encoding data into jpeg */
	private  JPEGParam  params;
	private  boolean    packedDataFlag;
	private  OutputStream   outStream;
	private  Raster    dataSrc;


	private static final Class OutputStreamClass = OutputStream.class;
	// native method that will initialize the method id's for all calls back to java.
	private native void initEncoder( Class OutputStreamClass);
	/** 
	 * Native method that is the only interface to the JPEG library that is loaded
	 * when the class is invoked.  
	 */
	private synchronized native void writeByteJPEGStream(byte[] data,  
											JPEGParam p, 
											OutputStream dest);
	/** 
	 * Native method that is the only interface to the JPEG library that is loaded
	 * when the class is invoked.  
	 */
	private synchronized native void writeShortJPEGStream(short[] data,  
											JPEGParam p, 
											OutputStream dest);

	/** 
	 * Native method that is the only interface to the JPEG library that is loaded
	 * when the class is invoked.  
	 */
	private synchronized native void writeIntJPEGStream(int[] data,  
											JPEGParam p, 
											OutputStream dest,
											boolean packedData);
	/**
	 * static method to load the JPEG lib when the class is constructed. 
	 */
	static { 
		try {
		    java.security.AccessController.beginPrivileged();
			System.loadLibrary("jpeg" );
			} 
		finally {
		    java.security.AccessController.endPrivileged();
		}
	}

	/**
	 * Constructs a JPEGImageEncoder using a Raster dataa source and destination
	 * OutputStream.
	 * @param src - the Raster containing the image data to be encoded.
	 * @param dest - destination of the encoded data.
	 */
	public JPEGImageEncoder(Raster src, OutputStream dest) {
		params = new JPEGParam();
		outStream = dest;
		dataSrc = src;
		packedDataFlag = false;
		initEncoder( OutputStreamClass);
	}
	/**
	 * Constructs a JPEGImageEncoder specifying the BufferedImage data source 
	 * and destination OutputStream.
	 * @param src - the BufferedImage containing the data to be encoded
	 * @param dest - destination of the encoded data.
	 */
	public JPEGImageEncoder(BufferedImage src, OutputStream dest) {
		dataSrc = src.getRaster();
		outStream = dest ;
		params = new JPEGParam();
		if( (src.getType() == BufferedImage.TYPE_INT_ARGB_PRE )||
			src.isAlphaPremultiplied() )
			packedDataFlag = true;
		else
			packedDataFlag = true;
			
		initEncoder( OutputStreamClass);
	}
	/**
	 * Constructor to be used when tables only 
	 * data is being written.
	 * @param dest - destination of the encoded data.
	 */
	public JPEGImageEncoder( OutputStream dest ){
		outStream = dest;
		params = new JPEGParam();
		packedDataFlag = false;
		initEncoder( OutputStreamClass);
	}


	/**
	 * Encodes the data that is encapsulated by the Raster or BufferedImage  according
	 * to the settings of the JPEGParam, and writes the information to the OutputStream.
	 * Used for writing Interchange jpeg streams ( standard jpeg/jfif files ) and 
	 * abbreviated image streams.  Note to write the image data in an abbrevieated 
	 * JPEG stream set the value of JPEGParam.write_tables = false.  This will suppress
	 * writing of the Quantization and Huffman tables.
	 * @return void
	 * @exception - throws IOException if there are errors writing the encoded data.
	 */
	public void encode() throws IOException {  
		if( params.getwriteImage() == false )
			writeByteJPEGStream (null, params, outStream ) ;		

		else{

			if(dataSrc.getSampleModel().getDataType() == DataBuffer.BYTE_DATA  ){
				writeByteJPEGStream (
				  ((DataBufferByte)dataSrc.getDataBuffer()).getData(),
									 params, outStream ) ;
			}
			else if(dataSrc.getSampleModel().getDataType() == DataBuffer.SHORT_DATA  ){
				writeShortJPEGStream (
				  ((DataBufferShort)dataSrc.getDataBuffer()).getData(),
									 params, outStream ) ;
			}
			else if(dataSrc.getSampleModel().getDataType() == DataBuffer.INT_DATA  ){
				if( packedDataFlag == true ){
					writeIntJPEGStream (
						   ((DataBufferInt)dataSrc.getDataBuffer()).getData(),
									 params, outStream, true ) ;
				}
			else
				{
				writeIntJPEGStream (
				  ((DataBufferInt)dataSrc.getDataBuffer()).getData(),
									 params, outStream, false ) ;
				}
			}
		}
	}
  
	/**
	 * getJPEGParam()- returns the current JPEGParam data member.
	 * @return params - the current JPEGParam
	 */
	public JPEGParam	getJPEGParam() {return params ;}
   
	/**
	 * Allows users to set the JPEGParam object that contains the information 
	 * used to encode image data.
	 * @param p - new value assigned to the params member.
	 * @return void
	 * @exception none
	 */
	public void setJPEGParam( JPEGParam p) {params = p;}

}





