/*
 * @(#)JPEGQTable.java	1.3 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image.codec;

/** Class to encapsulate the JPEG quantization tables */
public class JPEGQTable implements Cloneable {

	/** Quantization step for each coefficient in zig-zag order */
	private int quantval[];

	/** The number of coefficients in a DCT block */
	private static final byte _QTABLESIZE = 64;

	/** The zig-zag order of coefficients. Used to convert unordered tables */
	private static final byte zz[] = { 
			0,1,8,9,2,3,10,17,24,32,15,18,11,4,5,12,19,26,33,40,48,41,34,27,20,
			13,6,7,14,21,28,35,42,49,56,57,50,43,36,29,22,15,23,30,37,44,51,58,
			59,52,45,38,31,39,46,53,60,61,62,63 };

	/** This is the sample luminance quantization table given in the JPEG spec section K.1,
  	  * but expressed in zigzag order. The spec says that the values given produce "good" quality, and
	  * when divided by 2, "very good" quality.
	  */
	protected static final int std_luminance_quant_tbl[] = {
		16,  11,  12,  14,  12,  10,  16,  14,
		13,  14,  18,  17,  16,  19,  24,  40,
		26,  24,  22,  22,  24,  49,  35,  37,
		29,  40,  58,  51,  61,  60,  57,  51,
		56,  55,  64,  72,  92,  78,  64,  68,
		87,  69,  55,  56,  80, 109,  81,  87,
		95,  98, 103, 104, 103,  62,  77, 113,
		121, 112, 100, 120,  92, 101, 103,  99
	};

	/** This is the sample luminance quantization table given in the JPEG spec section K.1,
	  * but expressed in zigzag order. The spec says that the values given produce "good" quality, and
	  * when divided by 2, "very good" quality.
	  */
	protected static final int std_chrominance_quant_tbl[] = {
		17,  18,  18,  24,  21,  24,  47,  26,
		26,  47,  99,  66,  56,  66,  99,  99,
		99,  99,  99,  99,  99,  99,  99,  99,
		99,  99,  99,  99,  99,  99,  99,  99,
		99,  99,  99,  99,  99,  99,  99,  99,
		99,  99,  99,  99,  99,  99,  99,  99,
		99,  99,  99,  99,  99,  99,  99,  99,
		99,  99,  99,  99,  99,  99,  99,  99
	};

  /** Constructs an empty quantization table */
	public JPEGQTable()
	  {
		quantval = new int[_QTABLESIZE];

		}
  /** Constructs an quantization table with the array that was passed.
   *  @param  table - the new quantization table.
   */
	public JPEGQTable( int table[])
	  {
		quantval = new int[_QTABLESIZE];
		if ( table.length != _QTABLESIZE ) {
			throw new IllegalArgumentException( "Quantization table is the wrong size.");
			}
		else {
			System.arraycopy( table, 0, quantval, 0, _QTABLESIZE );
			}

		}
  /** Sets a new quantization table with the array that was passed.
   *  @param int[] table - The new quantization table
   *  @Exception IllegalArgumentException - thrown if the table size is not
   *     equal to the correct length of 64 values.
   */
	public void setTable( int table[])
	  {
		quantval = new int[_QTABLESIZE];
		if ( table.length != _QTABLESIZE ) {
			throw new IllegalArgumentException( "Quantization table is the wrong size.");
			}
		else {
			System.arraycopy( table, 0, quantval, 0, _QTABLESIZE );
			}

		}
	/** Puts an array of qantization steps in zig-zag order and copies
         *  them to the Q table.
         *  @param int[] table - The new quantization table
         *  @return void.
         *  @Exception IllegalArgumentException - thrown if the table size is
         *      not equal to the correct length of 64 values.
         */
	public void setUnorderedTable( int table[] ) {
		if ( table.length != _QTABLESIZE ) {
			throw new IllegalArgumentException( "Quantization table is the wrong size.");
			}
		else {
		   // Reorder coefficients
		   int[] ordered = new int[_QTABLESIZE];
		   for ( int i=0; i<_QTABLESIZE; i++ ){
			   ordered[i] = table[zz[i]];
			   }
		   quantval = ordered;
			}
		}

	/** Returns the current quantization table as an array of ints in
         *  zig zag order.
         *  @return The quantization table.
         */
	public int[] getTable()
		{ 
		int[] table = new int[_QTABLESIZE];
		System.arraycopy( quantval, 0, table, 0, _QTABLESIZE );
		return table;
		}

	/**
	 * The values in the Quantization table are multiplied by
         * scale_factor/100 and then clamped to the range 1..65535
	 * (or to 1..255 if force_baseline is TRUE).
	 * @param int factor - the factor by which the table will be scaled
	 * @param boolean  force_baseline - if true the scaling will be clamped
	 * to a range of 1.. 255
	 */

	public void scaleTable( int factor, boolean force_baseline ) {
		long holder;
		
		for (int i=0; i<_QTABLESIZE; i++ ) {
			holder = ((long)quantval[i] * factor + 50L) / 100L;

			/* limit to valid range */
			if (holder <= 0L) holder = 1L;
			if (holder > 32767L) holder = 32767L; /* Max quantizer needed for 12 bits */
			if (force_baseline && holder > 255L ) holder = 255L;

			quantval[i] = (int)holder;
			}
		}

	/** Clones this table. */
    public synchronized Object clone() {
	try { 
	    JPEGQTable t = (JPEGQTable)super.clone();
	    t.quantval = new int[_QTABLESIZE];
	    System.arraycopy(quantval, 0, t.quantval, 0, _QTABLESIZE);
	    return t;
	} catch (CloneNotSupportedException e) { 
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
    }

	}
