/*
 * @(#)ParameterBlock.java	1.3 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.image.renderable;
import java.awt.image.RenderedImage;
import java.io.Serializable;
import java.util.Vector;

/**
 * A ParameterBlock encapsulates all the information about sources and
 * parameters (expressed as base types or Objects) required by a
 * RenderableImageOp and other future classes that manipulate chains of
 * imaging operators.
 *
 * <p> Although it is possible to place arbitrary objects in the
 * source Vector, users of this class may impose semantic constraints
 * such as requiring all sources to be RenderedImages or
 * RenderableImage.  ParameterBlock itself is merely a container and
 * performs no checking on source or parameter types.
 *
 * <p> Note that all get() and set() methods operate on references.
 * Therefore, one must be careful not to share references between
 * ParameterBlocks when this is inappropriate.  For example, to create
 * a new ParameterBlock that is equal to an old one except for an
 * added source, one might be tempted to write:
 *
 * <pre>
 * ParameterBlock addSource(ParameterBlock pb, RenderableImage im) {
 *     ParameterBlock pb1 = new ParameterBlock(pb.getSources());
 *     pb1.addSource(im);
 *     return pb1;
 * }
 * </pre>
 *
 * <p> This code will have the side effect of altering the original
 * ParameterBlock, since the getSources() operation returned a reference
 * to its source Vector.  Both pb and pb1 share their source Vector,
 * and a change in either is visible to both.
 *
 * <p> A correct way to write the addSource function is to clone
 * the source Vector:
 *
 * <pre>
 * ParameterBlock addSource (ParameterBlock pb, RenderableImage im) {
 *     ParameterBlock pb1 = new ParameterBlock(pb.getSources().clone());
 *     pb1.addSource(im);
 *     return pb1;
 * }
 * </pre>
 *
 * <p> The clone() method of ParameterBlock has been defined to
 * perform a clone of both the source and parameter Vectors for
 * this reason.  A standard, shallow clone is available as
 * shallowClone().
 *
 * <p> The addSource() and add() methods are defined to return 'this'
 * after adding their argument.  This allows use of syntax like:
 *
 * <pre>
 * ParameterBlock pb = new ParameterBlock();
 * op = new RenderableImageOp("operation", pb.add(arg1).add(arg2));
 * </pre>
 *
 */
public class ParameterBlock implements Cloneable, Serializable {
    /** A Vector of sources, stored as arbitrary Objects. */
    protected Vector sources = new Vector();
    
    /** A Vector of non-source parameters, stored as arbitrary Objects. */
    protected Vector parameters = new Vector();

    /** A dummy constructor. */
    public ParameterBlock() {}

    /** Constructs a ParameterBlock with a given Vector of sources. */
    public ParameterBlock(Vector sources) {
        setSources(sources);
    }
    
    /**
     * Constructs a ParameterBlock with a given Vector of sources and
     * Vector of parameters.
     */
    public ParameterBlock(Vector sources, Vector parameters) {
        setSources(sources);
        setParameters(parameters);
    }
    
    /**
     * Creates a shallow copy of a ParameterBlock.  The source and
     * parameter Vectors are copied by reference -- additions or
     * changes will be visible to both versions.
     *
     * @return an Object clone of the ParameterBlock.
     */
    public Object shallowClone() {
        try {
            return super.clone();
        } catch (Exception e) {
            // We can't be here since we implement Cloneable.
            return null;
        }
    }

    /**
     * Creates a copy of a ParameterBlock.  The source and parameter
     * Vectors are cloned, but the actual sources and parameters are
     * copied by reference.  This allows modifications to the order
     * and number of sources and parameters in the clone to be invisible
     * to the original ParameterBlock.  Changes to the shared sources or
     * parameters themselves will still be visible.
     *
     * @return an Object clone of the ParameterBlock.
     */
    public Object clone() {
        ParameterBlock theClone;

        try {
            theClone = (ParameterBlock) super.clone();
        } catch (Exception e) {
            // We can't be here since we implement Cloneable.
            return null;
        }

        theClone.setSources((Vector) sources.clone());
        theClone.setParameters((Vector) parameters.clone());
        return (Object) theClone;
    }

    /**
     * Adds an image to end of the list of sources.  The image is
     * stored as an object in order to allow new node types in the
     * future.
     *
     * @param obj an image object to be stored in the source list.
     */
    public ParameterBlock addSource(Object obj) {
        sources.addElement(obj);
        return this;
    }
    
    /**
     * Returns a source as a general Object.  The caller must cast it into
     * an appropriate type.
     *
     * @param index the index of the source to be returned.
     */
    public Object getSource(int index) {
        return sources.elementAt(index);
    }
    
    /**
     * A convenience method to return a source as a RenderedImage.
     * An exception will be thrown if the sources is not a RenderedImage.
     *
     * @param index the index of the source to be returned.
     */
    public RenderedImage getRenderedSource(int index) {
        return (RenderedImage) sources.elementAt(index);
    }
    
    /**
     * A convenience method to return a source as a RenderableImage.
     * An exception will be thrown if the sources is not a RenderableImage.
     *
     * @param index the index of the source to be returned.
     */
    public RenderableImage getRenderableSource(int index) {
        return (RenderableImage) sources.elementAt(index);
    }

    /** Returns the number of source images. */
    public int getNumSources() {
        return sources.size();
    }
    
    /** Returns the entire Vector of sources. */
    public Vector getSources() {
        return sources;
    }
    
    /** Sets the entire Vector of sources to a given Vector. */
    public void setSources(Vector sources) {
        this.sources = sources;
    }
    
    /** Clears the list of source images. */
    public void removeSources() {
        sources = new Vector();
    }
    
    /** Returns the number of parameters (not including source images). */
    public int getNumParameters() {
        return parameters.size();
    }
    
    /** Returns the entire Vector of parameters. */
    public Vector getParameters() {
        return parameters;
    }
    
    /** Sets the entire Vector of parameters to a given Vector. */
    public void setParameters(Vector parameters) {
        this.parameters = parameters;
    }
    
    /** Clears the list of parameters. */
    public void removeParameters() {
        parameters = new Vector();
    }
    
    /** Adds a byte to the list of parameters. */
    public ParameterBlock add(byte b) {
        parameters.addElement(new Byte(b));
        return this;
    }
    
    /** Adds a char to the list of parameters. */
    public ParameterBlock add(char c) {
        parameters.addElement(new Character(c));
        return this;
    }
    
    /** Adds a short to the list of parameters. */
    public ParameterBlock add(short s) {
        parameters.addElement(new Short(s));
        return this;
    }
    
    /** Adds a int to the list of parameters. */
    public ParameterBlock add(int i) {
        parameters.addElement(new Integer(i));
        return this;
    }
    
    /** Adds a long to the list of parameters. */
    public ParameterBlock add(long l) {
        parameters.addElement(new Long(l));
        return this;
    }

    /** Adds a float to the list of parameters. */
    public ParameterBlock add(float f) {
        parameters.addElement(new Float(f));
        return this;
    }

    /** Adds a double to the list of parameters. */
    public ParameterBlock add(double d) {
        parameters.addElement(new Double(d));
        return this;
    }

    /** Adds an object to the list of parameters. */
    public ParameterBlock add(Object obj) {
        parameters.addElement(obj);
        return this;
    }

    /**
     * Gets a parameter as an object.  Objects belonging to a base type
     * (e.g., int) will be returned as a member of the corresponding class
     * type (e.g., Integer).
     */
    public Object getObjectParameter(int index) {
        return parameters.elementAt(index);
    }

    /**
     * A convenience class to return a parameter as a byte.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public byte getByteParameter(int index) {
        return ((Byte) parameters.elementAt(index)).byteValue();
    }

    /**
     * A convenience class to return a parameter as a char.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public char getCharParameter(int index) {
        return ((Character) parameters.elementAt(index)).charValue();
    }
    
    /**
     * A convenience class to return a parameter as a short.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public short getShortParameter(int index) {
        return ((Short) parameters.elementAt(index)).shortValue();
    }

    /**
     * A convenience class to return a parameter as an int.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public int getIntParameter(int index) {
        return ((Integer) parameters.elementAt(index)).intValue();
    }

    /**
     * A convenience class to return a parameter as a long.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public long getLongParameter(int index) {
        return ((Long) parameters.elementAt(index)).longValue();
    }

    /**
     * A convenience class to return a parameter as a float.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public float getFloatParameter(int index) {
        return ((Float) parameters.elementAt(index)).floatValue();
    }

    /**
     * A convenience class to return a parameter as a double.
     * An exception will be thrown if the parameter is of a different type.
     *
     * @param index the index of the parameter to be returned.
     */
    public double getDoubleParameter(int index) {
        return ((Double) parameters.elementAt(index)).doubleValue();
    }

    /**
     * Returns an array of Class objects describing the types
     * of the parameters.
     */
    public Class [] getParamClasses() {
        int numParams = getNumParameters();
        Class [] classes = new Class[numParams];
        int i;

        for (i = 0; i < numParams; i++) {
            Object obj = getObjectParameter(i);
            if (obj instanceof Byte) {
              classes[i] = byte.class;
            } else if (obj instanceof Character) {
              classes[i] = char.class;
            } else if (obj instanceof Short) {
              classes[i] = short.class;
            } else if (obj instanceof Integer) {
              classes[i] = int.class;
            } else if (obj instanceof Long) {
              classes[i] = long.class;
            } else if (obj instanceof Float) {
              classes[i] = float.class;
            } else if (obj instanceof Double) {
              classes[i] = double.class;
            } else {
              classes[i] = obj.getClass();
            }
        }
        
        return classes;
    }
}
