/*
 * @(#)RenderContext.java	1.3 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* ********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/

package java.awt.image.renderable;
import java.util.*;
import java.awt.geom.*;
import java.awt.*;
import java.awt.image.*;

/**
 * A RenderContext encapsulates the information needed to produce a
 * specific rendering from a RenderableImage.  It contains the area to
 * be rendered specified in rendering-independent terms, the
 * resolution at which the rendering is to be performed, and hints
 * used to control the rendering process.
 * 
 * Users create RenderContexts and pass them to the RenderableImage
 * via the getImage method.  Most of the methods of RenderContexts are
 * not meant to used by users, but by the RenderableImage and operator
 * classed which it may be passed to.
 *
 * The AffineTransform and Hashtable parameters passed into and out of
 * this class are cloned.  The Shape parameters are not necessarily
 * cloneable and are therefore only reference copied.  Changing
 * instances of Shape that are also in use by instances of
 * RenderContext may have undesired side effects.
 */
public class RenderContext implements Cloneable {
    
    /** Table of hints.  May be null. */
    Hashtable hints;
    
    /**
     * Table of hints that have been observed (that is, performed).
     * May be null.
     */
    Hashtable hintsObserved;
    
    /** Transform to convert user coordinates to device coordinates.  */
    AffineTransform usr2dev;
    
    /** The area of interest.  May be null. */
    Shape aoi;
    
    // Various constructors that allow different levels of
    // specificity. If the Shape is missing the whole renderable area
    // is assumed. If hints is missing no hints are assumed.  

    /**
     * Constructs a RenderContext with a given transform.
     * The area of interest is supplied as a Shape,
     * and the rendering hints are supplied as a Hashtable.
     *
     * @param usr2dev an AffineTransform.
     * @param aoi a Shape representing the area of interest.
     * @param hints a Hashtable containing rendering hints.
     */
    public RenderContext(AffineTransform usr2dev,
                         Shape aoi,
                         Hashtable hints) {
        if (hints != null) {
            this.hints = (Hashtable)hints.clone();
            this.hintsObserved = new Hashtable();
        } else {
            this.hints = null;
            this.hintsObserved = null;
        }
        this.aoi = aoi;
    
        this.usr2dev = (AffineTransform)usr2dev.clone();
    }
    
    /**
     * Constructs a RenderContext with a given transform.
     * The area of interest is taken to be the entire renderable area.
     * No rendering hints are used.
     *
     * @param usr2dev an AffineTransform.
     */
    public RenderContext(AffineTransform usr2dev) {
        this(usr2dev, null, null);
    }
    
    /**
     * Constructs a RenderContext with a given transform and rendering hints.
     * The area of interest is taken to be the entire renderable area.
     *
     * @param usr2dev an AffineTransform.
     * @param hints a Hashtable containing rendering hints.
     */
    public RenderContext(AffineTransform usr2dev, Hashtable hints) {
        this(usr2dev, null, hints);
    }
    
    /**
     * Constructs a RenderContext with a given transform and area of interest.
     * The area of interest is supplied as a Shape.
     * No rendering hints are used.
     *
     * @param usr2dev an AffineTransform.
     * @param aoi a Shape representing the area of interest.
     */
    public RenderContext(AffineTransform usr2dev, Shape aoi) {
        this(usr2dev, aoi, null);
    }
    
    /**
     * Returns an Enumeration of the hint keys.
     *
     * @return returns an enumeration of the keys in the
     *         hints Hashtable, or null if no hints are present.
     */
    public Enumeration getKeys() {
        if (hints != null) {
            return hints.keys();
        } else {
            return null;
        }
    }
    
    /**
     * Searches for a key in the Hashtable of rendering hints.
     * If the key is found, the associated value is returned.
     * The key and value are also copied into the hintsObserved
     * Hashtable.
     *
     * @param key the hint's key, as an Object.
     * @return the Object value associated with the key.
     */
    public Object observeHint(Object key) {
        Object h;
        if (hints != null) {
            h = hints.get(key);
            if (h != null) {
                hintsObserved.put(key, h);
            }
        } else {
            h = null;
        }
        return h;
    }
    
    /**
     * Returns the value of a hint from the hints Hashtable.
     * If there are no hints, or the desired key is not present,
     * null is returned.
     *
     * @param key the hint's key, as an Object.
     * @return the hint associated with the key, or null.
     */
    public Object getHint(Object key) {
        if (hints != null) {
            return hints.get(key);
        } else {
            return null;
        }
    }
    
    /**
     * Sets a hint in the hints Hashtable.
     *
     * @param key the hint's key, as an Object.
     * @param value the value to assign to the hint.
     */
    public void setHint(Object key, Object value) {
        if (hints != null) {
            hints.put(key, value);
        }
    }

    /**
     * Returns a copy of the Hashtable of hints that have been observed 
     * by this renderContext.  The hint keys and values are copied
     * by reference.
     *
     * @return a Hashtable containing the hints observed.
     */
    public Hashtable getHintsObserved() {
        if (hintsObserved != null) {
            return (Hashtable)hintsObserved.clone();
        } else {
            return null;
        }
    }
    
    /**
     * Removes a hint from the hintsObserved Hashtable.
     *
     * @param key The key for the hint to be removed from
     *        the observed hints table.
     */
    public void unobserveHint(Object key) {
        if (hintsObserved != null) {
            hintsObserved.remove(key);
        }
    }
    
    /**
     * Adds hints from a supplied Hashtable to the Hashtable of
     * observed hints.  The hints will also be copied into the hints
     * hashtable.  If the current hints tables are null, new tables
     * will be created.  The tables will each be copies of the
     * supplied table, because all the hints will be considered to
     * have been observed.
     *
     * @param sourceHints The source hints to add to this RenderContext's
     *        hintsObserved Hashtable.
     */
    public void addHintsObserved(Hashtable sourceHints) {
        // Do nothing if source is null
        if (sourceHints == null) {
            return;
        }

        if (hintsObserved != null) { // hints will also be non-null.
            Object key, value;
            Enumeration keys = sourceHints.keys();
            while (keys.hasMoreElements()) {
                try {
                    key = keys.nextElement();
                    value = sourceHints.get(key);
                
                    // Add the hint to the regular hints table.
                    hints.put(key,value);
                    // Add the hint to the observed table.
                    hintsObserved.put(key,value);
                } catch (NoSuchElementException e) {
                    // No need to do anything. This shouldn't
                    // happen because keys comes from sourceHints.
                } catch (NullPointerException e) {
                    // No need to do anything
                }
            }
        } else {
            // The only to get here is if the hintsObserved table is null.
            // In that case, make two copies of the source, one for the 
            // observed hints table, one for the regular hints table.

            hints = (Hashtable)sourceHints.clone();
            hintsObserved = (Hashtable)sourceHints.clone();
        }
    }
    
    /**
     * Adds the observed hints from another RenderContext to this
     * RenderContext's observed hints table.
     * 
     * @param rc The RenderContext from which to get the observed hints.
     */
    public void addHintsObserved(RenderContext rc) {
        if (rc.hints != null && rc != this) {
            addHintsObserved(rc.hintsObserved);
        }
    }
    
    /**
     * Sets hintsObserved to a new table. This is accomplished by
     * cloning the input parameter.  Any current observed hints are
     * removed from the hints table and discarded.  All entries in the
     * new observed hints table are added to the hints table.
     *
     * @param sourceObserved The source of observedHints.
     */
    public void setHintsObserved(Hashtable sourceObserved) {
        if (hintsObserved != null) {
            // Run through existing observed table and remove items
            // from the hints table
            Enumeration keys = hintsObserved.keys();
            while(keys.hasMoreElements()) {
                Object key = keys.nextElement();
                hints.remove(key);
            }
            
            if (sourceObserved == null) {
                hintsObserved = null;
                return;
            }

            // replace with new observed table
            hintsObserved = (Hashtable)sourceObserved.clone();

            // Make sure all new items are in the hints table
            Object value;
            keys = hintsObserved.keys();
            while (keys.hasMoreElements()) {
                Object key = keys.nextElement();
                value = hintsObserved.get(key);
                hints.put(key,value);
            }
        } else {
            if (sourceObserved == null) {
                hintsObserved = null;
                return;
            }
            // If there weren't any to start with, need to copy
            // into both locations
            hintsObserved = (Hashtable)sourceObserved.clone();
            hints = (Hashtable)sourceObserved.clone();
        }
    }
        
    /**
     * Sets the current user-to-device AffineTransform contained
     * in the RenderContext to a given transform.
     *
     * @param newTransform the new AffineTransform.
     */
    public void setTransform(AffineTransform newTransform) {
        usr2dev = (AffineTransform)newTransform.clone();
    }
        
    /**
     * Modifies the current user-to-device transform by prepending another
     * transform.  In matrix notation the operation is:
     * <pre>
     * [this] = [modTransform] x [this]
     * </pre>
     *
     * @param modTransform the AffineTransform to prepend to the
     *        current usr2dev transform.
     */
    public void transform(AffineTransform modTransform) {
        usr2dev.preConcatenate(modTransform);                
    }
    
    /**
     * Gets the current user-to-device AffineTransform.
     *
     * @return A reference to the current AffineTransform.
     */
    public AffineTransform getTransform() {
        return (AffineTransform)usr2dev.clone();
    }

    /**
     * Sets the current area of interest.  The old area is discarded.
     *
     * @param newAoi The new area of interest.
     */
    public void setAreaOfInterest(Shape newAoi) {
        aoi = newAoi;
    }

    /**
     * Gets the ares of interest currently contained in the
     * RenderContext.
     * 
     * @return a reference to the area of interest of the RenderContext, 
     *         or null if none is specified.
     */
    public Shape getAreaOfInterest() {
        return aoi;
    }

    /**
     * Makes a copy of a RenderContext. The area of interest is copied
     * by reference.  The usr2dev Transform and hints tables are
     * actually cloned.
     *
     * @return The new cloned RenderContext.
     */
    public Object clone() {
        RenderContext newRenderContext = new RenderContext(usr2dev,
                                                           aoi, hints);
        if (hintsObserved != null) {
            newRenderContext.hintsObserved = (Hashtable)hintsObserved.clone();
        }
        return newRenderContext;
    }
}


