/*
 * @(#)PageFormat.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.awt.print;

import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

/**
 * A PageFormat instance describes the size and orientation
 * of a page to be printed.
 */
public class PageFormat implements Cloneable
{

 /* Class Constants */
	
    /**
     * These orientation constants can be passed to
     * setOrientation(). The following descriptions
     * describe the orientations as they apply to
     * a coordinate system useful for reading text,
     * that is where x runs in the direction of
     * Roman text and y runs in the direction from
     * the first line to the second.
     *
     * LANDSCAPE
     *  The origin is at the bottom left of the paper with
     *  x running bottom to top and y running left to right.
     *  Note that this it not the Macintosh landscape but
     *  is the Window's and PostScript landscape.
     *
     * PORTRAIT
     *  The origin is at the top left of the paper with
     *  x running to the right and y running down the
     *  paper.
     *
     * REVERSE_LANDSCAPE
     *  The origin is in the top right of the paper with x
     *  running top to bottom and y running right to left.
     *  Note that this is the Macintosh landscape.
     */
    public static final int LANDSCAPE = 0;
    public static final int PORTRAIT = 1;
    public static final int REVERSE_LANDSCAPE = 2;

 /* Instance Variables */
 
    /**
     * A description of the physical piece of paper.
     */
    private Paper mPaper;
 	
    /**
     * The orientation of the current page. This will be
     * one of the constants: PORTRIAT, LANDSCAPE, or
     * REVERSE_LANDSCAPE,
     */
    private int mOrientation = PORTRAIT;
	
 /* Constructors */
 	
    /**
     * Create a default, letter sized, portrait oriented
     * page format.
     */
    public PageFormat()
    {
	mPaper = new Paper();
    }

 /* Instance Methods */
 	
    /**
     * Make a copy of this PageFormat.
     */
    public Object clone()
    {
	PageFormat newPage;

	try { 
	    newPage = (PageFormat) super.clone();
	    newPage.mPaper = (Paper)mPaper.clone();

	} catch (CloneNotSupportedException e) {
	    e.printStackTrace();
	    newPage = null;	// should never happen.
	}

	return newPage;
    }
 	
    /**
     * Return a rectangle describing the size of
     * the paper to be drawn on. The rectangle
     * returned is in the oriented user space,
     * that is the rectangle has been adjusted
     * to take into account the page orientation.
     * If the paper size is letter but the orientation
     * is landscape, then this method returns a rectangle
     * that is wider than it is tall. The origin of
     * the returned rectangle is always (0, 0)
     */
    public Rectangle2D getPaperArea()
    {
	double width = mPaper.getWidth();
	double height = mPaper.getHeight();

	Point2D.Double widthHeight = new Point2D.Double(width, height);
	getTransform().transform(widthHeight, widthHeight);
    
	return new Rectangle2D.Double(0, 0, widthHeight.x, widthHeight.y);
    }
 	
    /**
     *	Return a rectangle describing the position
     *	and size of the imageable area of the page.
     *	This rectangle has been adjusted to take
     *	into account the page orientation. The
     *	top left of this rectangle is relative to
     *	the paper area.
     */
    public Rectangle2D getImageableArea()
    {

	Rectangle2D area = mPaper.getImageableArea();
	int orientation = getOrientation();

	if (orientation == LANDSCAPE || orientation == REVERSE_LANDSCAPE){
	    double x = area.getY();
	    double y = area.getX();
	    double width = area.getHeight();
	    double height = area.getWidth();

	    area = new Rectangle2D.Double(x, y, width, height);
	}

	return area;
    }			
 	
 	
    /**
     * Return a copy of the paper object associated
     * with this PageFormat.
     */
    public Paper getPaper()
    {
	return (Paper)mPaper.clone();
    }

    /**
     * Set the paper object for this PageFormat.
     */
     public void setPaper(Paper paper)
     {
	 mPaper = paper;
     }

    /**
     * Set the page orientation. 'orientation' must be
     * one of the constants: PORTRAIT, LANDSCAPE,
     * or REVERSE_LANDSCAPE.
     *
     * 11/07/97 Notes:
     *  Need to decide on the methods behavior when
     *  given an invalid 'orientation'.
     */
    public void setOrientation(int orientation)
    {
	if(0 <= orientation && orientation <= REVERSE_LANDSCAPE){
	    mOrientation = orientation;
	}else{
	    // What to do with an invalid parameter?
	}
    }
	
    /**
     * Return the orientation of this PageFormat.
     */
     public int getOrientation()
     {
	 return mOrientation;
     }

    /**
     * Return a transform that will translate user
     * space drawing to the requested orientation
     * orientation of the page, taking  into account
     * portait-landscape and horizontal-vertical flipping.
     *
     * 01/09/98: Should not return AffineTransform but instead
     * an array.
     */
    public AffineTransform getTransform()
    {
	AffineTransform transform = null;
	
	switch(mOrientation){
	
	case LANDSCAPE:
	    transform = new AffineTransform(0, -1, 1, 0,
					    0, mPaper.getHeight());
	    break;

	case PORTRAIT:
	    transform = new AffineTransform();
	    break;
		
	case REVERSE_LANDSCAPE:
	    transform = new AffineTransform(0, 1, -1, 0, 0, 0);
	    break;
	 
	default:
	    // should never get here. throw something?
	}

	return transform;
    }

}
