/*
 * @(#)JFileChooser.java	1.21 98/02/02
 * 
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */

package java.awt.swing.preview;

import java.awt.swing.*;
import java.awt.swing.plaf.*;
import java.io.File;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import java.awt.accessibility.*;

/**
 * A dialog that displays the files in a directory and lets the user 
 * select a file, browse to another directory, or specify a filename.
 * <p>
 * Helper classes:<br>
 * <ul>
 * <li>TypedFile: Used to return the selected file.
 * <li>FileType:  Used to define the filename extensions (file types)
 *                the user can select to limit the files that are
 *                displayed in the JDirectoryPane.
 * <li>JDirectoryPane: Displays the files in a given directory.
 * </ul>
 *
 * NOTE: This API is not complete. It will be changing past 0.7
 * PENDING(jeff) remove this comment when API is complete.
 * <p>
 * Warning: serialized objects of this class will not be compatible with
 * future swing releases.  The current serialization support is appropriate 
 * for short term storage or RMI between Swing1.0 applications.  It will
 * not be possible to load serialized Swing1.0 objects with future releases
 * of Swing.  The JDK1.2 release of Swing will be the compatibility
 * baseline for the serialized form of Swing objects.
 *
 * @version 1.21 02/02/98
 * @author Ray Ryan
 * @see FileType
 * @see TypedFile
 */
public class JFileChooser extends JComponent implements SwingConstants, Accessible {
    /**
     * The filename patterns that the user can select from the 
     * file-types dropdown box.
     */
    protected Vector choosableTypes = new Vector();

    /**
     * The label for the field in which the user enters the filename
     * or pattern of filenames to display.
     */
    protected String prompt = DefaultPrompt;
    /**
     * The label for the drop-down list that displays pre-defined
     * filename patterns, such as <pre>
     *    All Files (*.*)
     * </pre>
     */
    protected String typesTitle = DefaultTypesTitle;
    /**
     * The label for the combo box that shgws the current directory.
     */
    protected String locationTitle = DefaultLocationTitle;
    /**
     * The label for the OK button.
     */
    protected String okayTitle = DefaultOkayTitle;
    /**
     * The label for the cancel button.
     */
    protected String cancelTitle = DefaultCancelTitle;

    /**
     * Default values for the dialog properties.
     */
    // PENDING(jeff) - These are too win32 specific. Genericize.
    static protected String DefaultPrompt = "File Name:";
    static protected String DefaultTypesTitle = "Files of Type:";
    static protected String DefaultLocationTitle = "Look in:";
    static protected String DefaultOkayTitle = "OK";
    static protected String DefaultCancelTitle = "Cancel";

    /**
     * The command strings that are put into the action events
     * generated by the dialog.
     */
    protected String okayCommand = DefaultOkayCommand;
    protected String cancelCommand = DefaultCancelCommand;

    /**
     * Default values for the event-commands.
     */
    // DO NOT LOCALIZE THE FOLLOWING!
    static protected String DefaultOkayCommand = "OKAY";
    static protected String DefaultCancelCommand = "CANCEL";

    /**
     * The pane that displays the files in a directory.
     */
    protected JDirectoryPane directoryPane;
    /**
     * TBD...
     */
    protected Component accessory = null;

    /**
     * Creates a JFileChooser on the user's home directory.
     */
    public JFileChooser() {
        this((File)null);
    }

    /**
     * Creates a JFileChooser on the given path.
     *
     * @param path  a String giving the path to a file or directory
     */
    public JFileChooser(String path) {
        this(path == null ? null : new File(path));
    }


    /**
     * Creates a JFileChooser on the given File
     *
     * @param directory  a File object specifying the path to a file 
     *                   or directory
     */
    public JFileChooser(File directory) {
        super();
        directoryPane = new JDirectoryPane(directory);
        updateUI();
    }

    //
    // Various title strings
    //

    /**
     * Returns the label for the field in which the user enters the 
     * filename or pattern of filenames to display.
     *
     * @return the String containing the field label
     */
    public String getPrompt() {
        return prompt;
    }
    
    /**
     * Sets the label for the field in which the user enters the filename
     * or pattern of filenames to display.
     *
     * @param prompt the String to display 
     */
    public void setPrompt(String prompt) {
        String old = this.prompt;
        this.prompt = prompt;
        this.firePropertyChange("prompt", old, prompt);
    }
    
    /**
     * Returns the label for the drop-down list that displays pre-defined
     * filename patterns.
     *
     * @return the String containing the field label
     */
    public String getTypesTitle() {
        return typesTitle;
    }
    
    /**
     * Sets the label for the drop-down list that displays pre-defined
     * filename patterns.
     *
     * @param prompt the String to display 
     */
    public void setTypesTitle(String typesTitle) {
        String old = this.typesTitle;
        this.typesTitle = typesTitle;
        this.firePropertyChange("typesTitle", old, typesTitle);
    }
    
    /**
     * Returns the label for the combo box that shgws the current directory.
     *
     * @return the String containing the field label
     */
    public String getLocationTitle() {
        return locationTitle;
    }
    
    /**
     * Sets the label for the combo box that shgws the current directory.
     *
     * @param locationTitle  the String to display 
     */
    public void setLocationTitle(String locationTitle) {
        String old = this.locationTitle;
        this.locationTitle = locationTitle;
        this.firePropertyChange("locationTitle", old, locationTitle);
    }
    
    /**
     * Returns the label for the OK button.
     *
     * @return the String containing the button text
     */
    public String getOkayTitle() {
        return okayTitle;
    }
    
    /**
     * Sets the label for the OK button.
     *
     * @param okayTitle  the String to display 
     */
    public void setOkayTitle(String okayTitle) {
        String old = this.okayTitle;
        this.okayTitle = okayTitle;
        this.firePropertyChange("okayTitle", old, okayTitle);
    }
    
    /**
     * Returns the label for the cancel button.
     *
     * @return the String containing the button text
     */
    public String getCancelTitle() {
        return cancelTitle;
    }
    
    /**
     * Sets the label for the cancel button.
     *
     * @param cancelTitle  the String to display 
     */
    public void setCancelTitle(String cancelTitle) {
        String old = this.cancelTitle;
        this.cancelTitle = cancelTitle;
        this.firePropertyChange("cancelTitle", old, cancelTitle);
    }

    /**
     * Returns the directory pane that displays the files.
     *
     * @return the JDirectoryPane that handles the file display
     */
    public JDirectoryPane getDirectoryPane() {
        return directoryPane;
    }

    /**
     * Returns the file selected by the user.
     *
     * @return the TypedFile object encapsulating the selection
     * @see TypedFile
     */
    public TypedFile getSelectedFile() {
        return directoryPane.getSelectedFile();
    }

    /**
     * Called by the UI when the user accepts input.
     */
    public void performOkay() {
        directoryPane.fireActionPerformed(this.getOkayCommand());
    }

    /**
     * Called by the UI when the user aborts.
     */
    public void performCancel() {
        directoryPane.fireActionPerformed(this.getCancelCommand());
    }

    //
    // Command strings
    //

    /**
     * Returns the command included in the ActionEvent when the Okay button
     * is clicked
     *
     * @return the command String
     */
    public String getOkayCommand() {
        return okayCommand;
    }
    
    /**
     * Sets the command included in the ActionEvent when the Okay button
     * is clicked
     *
     * @param okayCommand  the command String
     */
    public void setOkayCommand(String okayCommand) {
        String old = this.okayCommand;
        this.okayCommand = okayCommand;
        this.firePropertyChange("okayCommand", old, okayCommand);
    }
    
    /**
     * Returns the command included in the ActionEvent when the Cancel button
     * is clicked or the dialog is closed.
     *
     * @return the command String
     */
    public String getCancelCommand() {
        return cancelCommand;
    }
    
    /**
     * Sets the command included in the ActionEvent when the Cancel button
     * is clicked or the dialog is closed.
     *
     * @param cancelCommand  the command String
     */
    public void setCancelCommand(String cancelCommand) {
        String old = this.cancelCommand;
        this.cancelCommand = cancelCommand;
        this.firePropertyChange("cancelCommand", old, cancelCommand);
    }

    //
    // Sessions
    //

    /**
     * Return 0 on Okay, -1 on Cancel or other abort
     */
    public int showDialog(Component parent) {
        Frame f = parent instanceof Frame ? (Frame) parent
            : (Frame)SwingUtilities.getAncestorOfClass(Frame.class, parent);

        JDialog dialog = new JDialog(f, "Save", true);
        Container contentPane = dialog.getContentPane();
        contentPane.setLayout(new BorderLayout());
        contentPane.add(this, BorderLayout.CENTER);

        dialog.pack();
        dialog.setLocationRelativeTo(parent);

        ModalListener l = new ModalListener(dialog);
        directoryPane.addActionListener(l);
        dialog.show();
        directoryPane.removeActionListener(l);
        return l.returnValue;
    }

    /**
     *
     */
    protected class ModalListener implements ActionListener {
        public int returnValue = -1;
        JDialog dialog;

        public ModalListener(JDialog d) { dialog = d; }     

        public void actionPerformed(ActionEvent e) {
            String command = e.getActionCommand();
            if (command != null) {
                if (command.equals(getOkayCommand())) {
                    returnValue = 0;
                    this.dialog.setVisible(false);
                }
                if (command.equals(getCancelCommand())) {
                    returnValue = -1;
                    this.dialog.setVisible(false);
                }
            }
        }
    }
                

    //
    // Accessory View
    //

    /**
     * Return the accessory component.
     *
     * @return this JFileChooser's accessory component, or null
     */
    public Component getAccessory() {
        return accessory;
    }

    /**
     * Set this JFileChooser's accessory component. If the old one was
     * listening for any events, make it stop. If the new one is a
     * PropertyChangeListener or an ActionListener, call the appropriate
     * add message on this JFileChooser.
     */
    public void setAccessory(Component c) {
        Component old = this.accessory;
        if (old instanceof PropertyChangeListener) {
            this.removePropertyChangeListener((PropertyChangeListener)old);
        }
        if (old instanceof ActionListener) {
            directoryPane.removeActionListener((ActionListener)old);
        }

        this.accessory = c;

        if (c instanceof PropertyChangeListener) {
            this.addPropertyChangeListener((PropertyChangeListener)c);
        }
        if (c instanceof ActionListener) {
            directoryPane.addActionListener((ActionListener)c);
        }

        this.firePropertyChange("accessory", old, c);
    }

    //
    // Types Available to the User
    //

    /**
     * Adds a type consisting of a single file extension to the set of
     * types the user can choose to filter by. 
     *
     * @param presentationName  a String containing the text to display
     *              in the list. For example: "All files (*.*)"
     * @param extension  a String that specifies the filename-extension
     *                   that is used to limit selections.  
     * @param icon  an Icon to display that indicates the file type
     *              (may be null)
     */
    public void addChoosableFileType(String presentationName,
                                     String extension,
                                     Icon icon) {
        FileType type = new FileType.
            ExtensionBased(presentationName, extension, icon);
        this.addChoosableFileType(type);
    }

    /**
     * Adds a type consisting of a set of file extenions to the set of
     * types the user can choose to filter by.
     *
     * @param presentationName  a String containing the text to display
     *              in the list. For example: "HTML files (*.htm, *.html)"
     * @param extension  an array of Strings that specifies the 
     *                   filename-extensions that limit selections.  
     * @param icon  an Icon to display that indicates the file type
     *              (may be null)
     */
    public void addChoosableFileType(String presentationName,
                                     String[] extensions,
                                     Icon icon) {
        FileType type = new FileType.
            ExtensionBased(presentationName, extensions, icon);
        this.addChoosableFileType(type);
    }

    /**
     * Adds the given type to the set of types the user can
     * choose to filter by.
     *
     * @param type  the FileType object that encapsulates the String
     *              that describes the filename-template and the
     *              filename extensions (file types) that are
     *              displayed.
     * @see FileType
     */
    public void addChoosableFileType(FileType type) {
        if (!choosableTypes.contains(type)) {
            choosableTypes.addElement(type);
        }
        if (!directoryPane.isKnownFileType(type)) {
            directoryPane.addKnownFileType(type);
        }
        // PENDING(jeff): No op!
        this.firePropertyChange("choosableFileTypes", null, null);
    }
        
    /**
     * Returns an enumeration of the FileTypes that the user
     * can choose to filter by.
     *
     * @return an enumeration of FileType objects
     * @see FileType
     * @see java.util.Enumeration
     */
    public Enumeration enumerateChoosableFileTypes() {
        return choosableTypes.elements();
    }

    /**
     * Returns a vector of the FileTypes that the user
     * can choose to filter by.
     *
     * @return an enumeration of FileType objects
     * @see FileType
     * @see java.util.Vector
     */
    public Vector getChoosableFileTypes() {
        return (Vector)choosableTypes.clone();
    }

    /**
     * Returns true if the specified FileType object is one of
     * the types that the user can choose to filter by.
     * 
     * @param t  the FileType object to check
     * @return true if the user can use that FileType object for
     *              filtering
     * @see FileType
     */
    public boolean isChoosableFileType(FileType t) {
        return choosableTypes.contains(t);
    }

    /**
     * Specifies the complete set of file types the user can
     * choose to filter by.
     *
     * @param type  an array of FileType objects that encapsulate the 
     *              String that describes a filename-template and the
     *              filename extensions (file types) that are
     *              displayed when that template is chosen.
     * @see FileType
     */
    public void setChoosableFileTypes(FileType types[]) {
        int count = types.length;
        Vector newTypes = this.choosableTypes = new Vector(count);

        newTypes.setSize(count);
        for (int i = 0; i < count; i++) {
            FileType t = types[i];
            newTypes.setElementAt(t, i);
        }
        // PENDING(jeff): No op!
        this.firePropertyChange("choosableFileTypes", null, null);
    }

    /**
     * Notification from the UIManager that the L&F has changed. 
     * Replaces the current UI object with the latest version from the 
     * UIManager.
     *
     * @see JComponent#updateUI
     */
    public void updateUI() {
        setUI((FileChooserUI)UIManager.getUI(this));
    }

    /**
     * Returns the name of the L&F class that renders this component.
     *
     * @return "FileChooserUI"
     * @see JComponent#getUIClassID
     * @see UIDefaults#getUI
     */
    public String getUIClassID() {
        return "FileChooserUI";
    }

/// 
/// Accessibility Support 
///

    /** 
     * Get the AccessibleContext of this object 
     *
     * @return the AccessibleContext of this object
     */
    public AccessibleContext getAccessibleContext() {
        if (accessibleContext == null) {
            accessibleContext = new AccessibleJFileChooser();
        }
        return accessibleContext;
    }

    /**
     * The class used to obtain the accessible role for this object.
     * <p>
     * Warning: serialized objects of this class will not be compatible with
     * future swing releases.  The current serialization support is appropriate
     * for short term storage or RMI between Swing1.0 applications.  It will
     * not be possible to load serialized Swing1.0 objects with future releases
     * of Swing.  The JDK1.2 release of Swing will be the compatibility
     * baseline for the serialized form of Swing objects.
     */
    protected class AccessibleJFileChooser extends AccessibleJComponent {

        /**
         * Get the role of this object.
         *
         * @return an instance of AccessibleRole describing the role of the 
         * object
         * @see AccessibleRole
         */
        public AccessibleRole getAccessibleRole() {
            return AccessibleRole.FILE_CHOOSER;
        }

    }  // AccessibleJFileChooser
}

