/*
 * @(#)BeanContextServicesSupport.java	1.4 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.beans.beancontext;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import java.util.TooManyListenersException;

import java.util.Locale;

/**
 * <p>
 * This helper class provides a utility implementation of the
 * java.beans.beancontext.BeanContextServices interface.
 * </p>
 * <p>
 * Since this class directly implements the BeanContextServices interface,
 * the class can, and is intended to be used either by subclassing this
 * implementation, or via ad-hoc delegation of an instance of this class
 * from another.
 * </p>
 *
 * @author Laurence P. G. Cable
 * @version 1.4
 * @since JDK1.2
 */

public class      BeanContextServicesSupport extends BeanContextSupport
       implements BeanContextServices {

    /**
     * <p>
     * Construct a BeanContextServicesSupport instance
     * </p>
     *
     * @param peer	The peer BeanContext we are supplying an implementation for, if null the this object is its own peer
     * @param lcle	The current Locale for this BeanContext.
     * @param dtime	The initial state, true if in design mode, false if runtime.
     * @param visible	The initial visibility.
     * 
     */

    public BeanContextServicesSupport(BeanContextServices peer, Locale lcle, boolean dTime, boolean visible) {
	super(peer, lcle, dTime, visible);
    }

    /**
     * Create an instance using the specified Locale and design mode.
     *
     * @param peer	The peer BeanContext we are supplying an implementation for, if null the this object is its own peer
     * @param lcle	The current Locale for this BeanContext.
     * @param dtime	The initial state, true if in design mode, false if runtime.
     */

    public BeanContextServicesSupport(BeanContextServices peer, Locale lcle, boolean dtime) {
	this (peer, lcle, dtime, true);
    }

    /**
     * Create an instance using the specified locale
     *
     * @param peer	The peer BeanContext we are supplying an implementation for, if null the this object is its own peer
     * @param lcle	The current Locale for this BeanContext.
     */

    public BeanContextServicesSupport(BeanContextServices peer, Locale lcle) {
	this (peer, lcle, false, true);
    }

    /**
     * Create an instance with a peer
     *
     * @param peer	The peer BeanContext we are supplying an implementation for, if null the this object is its own peer
     */

    public BeanContextServicesSupport(BeanContextServices peer) {
	this (peer, null, false, true);
    }

    /**
     * Create an instance that is not a delegate of another object
     */

    public BeanContextServicesSupport() {
	this (null, null, false, true);
    }

    /**
     * called by BeanContextSupport superclass during construction and 
     * deserialization to initialize subclass transient state.
     *
     * subclasses may envelope this method, but should not override it or
     * call it directly.
     */

    public void initialize() {
	super.initialize();

	services     = new HashMap(serializable + 1);
	bcsListeners = new ArrayList(1);
    }

    /**
     * @return the instance of BeanContext this object is providing the implemen
tation for.
     */
 
    public BeanContextServices getBeanContextServicesPeer() { 
	return (BeanContextServices)getBeanContextChildPeer();
    }

    /************************************************************************/

    /*
     * protected nested class containing per child information, an instance
     * of which is associated with each child in the "children" hashtable.
     * subclasses can extend this class to include their own per-child state.
     *
     * Note that this 'value' is serialized with the corresponding child 'key'
     * when the BeanContextSupport is serialized.
     */

    protected class BCSSChild extends BeanContextSupport.BCSChild  {

	/*
	 * private nested class to map serviceClass to Provider and requestors
	 * listeners.
	 */

	class BCSSCServiceClassRef {

	    // create an instance of a service ref

	    BCSSCServiceClassRef(Class sc, BeanContextServiceProvider bcsp) {
		super();
	
		serviceClass    = sc;
		serviceProvider = bcsp;
	    }

	    // add a requestor and assoc listener

	    void addRequestor(Object requestor, BeanContextServiceRevokedListener bcsrl) throws TooManyListenersException {
		BeanContextServiceRevokedListener cbcsrl = (BeanContextServiceRevokedListener)requestors.get(requestor);

		if (cbcsrl != null && !cbcsrl.equals(bcsrl))
		    throw new TooManyListenersException();

		requestors.put(requestor, bcsrl);
	    }

	    // remove a requestor

	    void removeRequestor(Object requestor) {
		requestors.remove(requestor);
	    }

	    // check a requestors listener

	    void checkRequestor(Object requestor, BeanContextServiceRevokedListener bcsrl) throws TooManyListenersException {
		BeanContextServiceRevokedListener cbcsrl = (BeanContextServiceRevokedListener)requestors.get(requestor);

		if (cbcsrl != null && !cbcsrl.equals(bcsrl))
		    throw new TooManyListenersException();
	    }

	    Iterator entries() { return requestors.entries().iterator(); }

	    boolean isEmpty() { return requestors.isEmpty(); }

	    Class getServiceClass() { return serviceClass; }

	    BeanContextServiceProvider getBeanContextServiceProvider() {
		return serviceProvider;
	    }

	    /*
	     * fields
	     */

	    Class				serviceClass;
	    BeanContextServiceProvider		serviceProvider;

	    HashMap				requestors = new HashMap(1);
	}

     	/*
	 *
	 */

	class BCSSCServiceRef {
	    BCSSCServiceRef(BCSSCServiceClassRef scref) {
		serviceClassRef = scref;
	    }

	    void addRef()  { refCnt++;        }
	    int  release() { return --refCnt; }

	    BCSSCServiceClassRef getServiceClassRef() { return serviceClassRef; }
	    /*
    	     * fields
	     */

	    BCSSCServiceClassRef serviceClassRef;
	    int			 refCnt	= 1;
	}

	BCSSChild(Object bcc)            { super(bcc); }

        // note usage of service per requestor, per service

	synchronized void usingService(Object requestor, Object service, Class serviceClass, BeanContextServiceProvider bcsp, BeanContextServiceRevokedListener bcsrl)  throws TooManyListenersException {
	
	    // first, process mapping from serviceClass to requestor(s)

	    BCSSCServiceClassRef serviceClassRef = null;

	    if (serviceClasses == null) 
		serviceClasses = new HashMap(1);
	    else
	        serviceClassRef = (BCSSCServiceClassRef)serviceClasses.get(serviceClass);

	    if (serviceClassRef == null) { // new service being used ...
	    	serviceClassRef = new BCSSCServiceClassRef(serviceClass, bcsp);
		serviceClasses.put(serviceClass, serviceClassRef);

		serviceClassRef.addRequestor(requestor, bcsrl);

	    } else { // existing service ...
		serviceClassRef.checkRequestor(requestor, bcsrl); // throws
	    }

	    // now handle mapping from requestor to service(s)

	    BCSSCServiceRef serviceRef = null;
	    Map		    services   = null;

	    if (serviceRequestors == null) {
		serviceRequestors = new HashMap(1);
	    } else {
		services = (Map)serviceRequestors.get(requestor);
	    }

	    if (services == null) {
		services = new HashMap(1);

		serviceRequestors.put(requestor, services);
	    } else 
		serviceRef = (BCSSCServiceRef)services.get(service);

	    if (serviceRef == null) {
		serviceRef = new BCSSCServiceRef(serviceClassRef);

		services.put(service, serviceRef);
	    } else {
		serviceRef.addRef();
	    }
	}

	// release a service reference

	synchronized void releaseService(Object requestor, Object service) {
	    if (serviceRequestors == null) return;

	    Map services = (Map)serviceRequestors.get(requestor);
	
	    if (services == null) return; // oops its not there anymore!

	    BCSSCServiceRef serviceRef = (BCSSCServiceRef)services.get(service);

	    if (serviceRef == null) return; // oops its not there anymore!
	    
	    BCSSCServiceClassRef serviceClassRef = serviceRef.getServiceClassRef();
	    BeanContextServiceProvider bcsp = serviceClassRef.getBeanContextServiceProvider();

	    bcsp.releaseService(BeanContextServicesSupport.this.getBeanContextServicesPeer(), requestor, service);

	    if (serviceRef.release() == 0) {

		services.remove(service);

		if (services.isEmpty()) {
		    serviceRequestors.remove(requestor);
		}

		if (serviceRequestors.isEmpty()) {
		    serviceRequestors = null;
		}

		serviceClassRef.removeRequestor(requestor);

		if (serviceClassRef.isEmpty()) {
		    serviceClasses.remove(serviceClassRef.getServiceClass());
		}

		if (serviceClasses.isEmpty())
		    serviceClasses = null;
	    }
	}

	// revoke a service

	synchronized void revokeService(Class serviceClass, boolean revokeNow) {
	    if (serviceClasses == null) return;

	    BCSSCServiceClassRef serviceClassRef = (BCSSCServiceClassRef)serviceClasses.get(serviceClass);

	    if (serviceClassRef == null) return;

	    Iterator i = serviceClassRef.entries();

	    BeanContextServiceRevokedEvent bcsre = new BeanContextServiceRevokedEvent(BeanContextServicesSupport.this.getBeanContextServicesPeer(), serviceClass, revokeNow);

	    while (i.hasNext()) {
		Map.Entry	    		  entry    = (Map.Entry)i.next();
		BeanContextServiceRevokedListener listener = (BeanContextServiceRevokedListener)entry.getValue();

		if (revokeNow) {
		    Object  requestor = entry.getKey();
		    Map     services  = (Map)serviceRequestors.get(requestor);

		    if (services != null) {
			Iterator i1 = services.entries().iterator();

			while (i1.hasNext()) {
			    Map.Entry       tmp        = (Map.Entry)i1.next();

			    BCSSCServiceRef serviceRef = (BCSSCServiceRef)tmp.getValue();
			    if (serviceRef.getServiceClassRef().equals(serviceClassRef)) {	
				i1.remove();
			    }
			}

			if (services.isEmpty()) {
			    serviceRequestors.remove(requestor);
			}
		    }

		    serviceClassRef.removeRequestor(requestor);
		}

		listener.serviceRevoked(bcsre);
	    }

	    if (revokeNow && serviceClasses != null) {
		serviceClasses.remove(serviceClass);

	        if (serviceClasses.isEmpty())
		    serviceClasses = null;
	    }

	    if (serviceRequestors != null && serviceRequestors.isEmpty())
		serviceRequestors = null;
	}

	// release all references for this child since it has been unnested.

	void cleanupReferences() {

	    if (serviceRequestors == null) return;

	    Iterator requestors = serviceRequestors.entries().iterator();

	    while(requestors.hasNext()) {
		Map.Entry 	     tmp       = (Map.Entry)requestors.next();
		Object 	             requestor = tmp.getKey();
		Iterator 	     services  = ((Map)tmp.getValue()).entries().iterator();

		serviceRequestors.remove(requestor);

		while (services.hasNext()) {
		    Map.Entry       entry   = (Map.Entry)services.next();
		    Object 	    service = entry.getKey();
		    BCSSCServiceRef sref    = (BCSSCServiceRef)entry.getValue();

		    BCSSCServiceClassRef       scref = sref.getServiceClassRef();
		
		    BeanContextServiceProvider bcsp  = scref.getBeanContextServiceProvider();

		    scref.removeRequestor(requestor);
		    services.remove();

		    while (sref.release() >= 0) {
			bcsp.releaseService(BeanContextServicesSupport.this.getBeanContextServicesPeer(), requestor, service);
		    }
		}
	    }

	    serviceRequestors = null;
	    serviceClasses    = null;
	}

	/*
 	 * fields 
	 */

	private transient HashMap	serviceClasses;	   
	private transient HashMap	serviceRequestors;
    }

    /**
     * <p>
     * Subclasses can override this method to insert their own subclass
     * of Child without having to override add() or the other Collection
     * methods that add children to the set.
     * </p>
     *
     * @param targetChild the child to create the Child on behalf of
     */

    protected BCSChild createBCSChild(Object targetChild) {
	return new BCSSChild(targetChild);
    }

    /************************************************************************/

	/**
	 * subclasses may subclass this nested class to add behaviors for
	 * each BeanContextServicesProvider.
	 */

	protected static class BCSSServiceProvider implements Serializable {

	    BCSSServiceProvider(Class sc, BeanContextServiceProvider bcsp) {
		super();
	
		serviceProvider = bcsp;
	    }

	    protected BeanContextServiceProvider getServiceProvider() {
		return serviceProvider;
	    }

	    /*
   	     * fields
	     */

	    protected BeanContextServiceProvider serviceProvider;
	}

    	/**
	 * subclasses can override this method to create new subclasses of
	 * BCSSServiceProvider without having to overrride addService() in
	 * order to instantiate.
	 */

	protected BCSSServiceProvider createBCSSServiceProvider(Class sc, BeanContextServiceProvider bcsp) {
	    return new BCSSServiceProvider(sc, bcsp);
	}

    /************************************************************************/

    /**
     * add a BeanContextServicesListener
     */

    public void addBeanContextServicesListener(BeanContextServicesListener bcsl) {
	synchronized(bcsListeners) {
	    if (bcsListeners.contains(bcsl))
		return;
	    else
	        bcsListeners.add(bcsl);
	}
    }

    /**
     * remove a BeanContextServicesListener
     */

    public void removeBeanContextServicesListener(BeanContextServicesListener bcsl) {
	synchronized(bcsListeners) {
	    if (!bcsListeners.contains(bcsl))
		return;
	    else
	        bcsListeners.remove(bcsl);
	}
    }

    /**
     * add a service
     */

    public synchronized boolean addService(Class serviceClass, BeanContextServiceProvider bcsp) {
	return addService(serviceClass, bcsp, true);
    }

    /**
     * add a service
     */

    protected synchronized boolean addService(Class serviceClass, BeanContextServiceProvider bcsp, boolean fireEvent) {
        if (services.containsKey(serviceClass))
	    return false;
	else {
	    services.put(serviceClass,  createBCSSServiceProvider(serviceClass, bcsp));

	    if (bcsp instanceof Serializable) serializable++;

	    if (!fireEvent) return true;


	    BeanContextServiceAvailableEvent bcssae = new BeanContextServiceAvailableEvent(getBeanContextServicesPeer(), serviceClass);

	    fireServiceAdded(bcssae);

	    Iterator i = children.keySet().iterator();

	    while (i.hasNext()) {
	        Object c = i.next();

	        if (c instanceof BeanContextServices) {
		    ((BeanContextServicesListener)c).serviceAvailable(bcssae);
	        }
	    }

	    return true;
	}
    }

    /**
     * remove a service
     */

    public synchronized void revokeService(Class serviceClass, BeanContextServiceProvider bcsp, boolean revokeCurrentServicesNow) {
	if (!services.containsKey(serviceClass)) return;

	BCSSServiceProvider bcsssp = (BCSSServiceProvider)services.get(serviceClass);
	    
	if (!bcsssp.getServiceProvider().equals(bcsp)) 
	    throw new IllegalArgumentException("service provider mismatch");

	services.remove(serviceClass);

        if (bcsp instanceof Serializable) serializable--;

	Iterator i = bcsChildren(); // get the BCSChild values.

	while (i.hasNext()) {
	    ((BCSSChild)i.next()).revokeService(serviceClass, revokeCurrentServicesNow);
	}

	fireServiceRevoked(serviceClass, revokeCurrentServicesNow);
    }

    /**
     * has a service, which may be delegated
     */

    public synchronized boolean hasService(Class serviceClass) {
	if (services.containsKey(serviceClass)) return true;

	BeanContextServices bcs = null;

	try {
	    bcs = (BeanContextServices)getBeanContext();
	} catch (ClassCastException cce) {
	    return false;
	}

	return bcs == null ? false : bcs.hasService(serviceClass);
    }

    /************************************************************************/

    /*
     * a nested subclass used to represent a proxy for a serviceClass delegated
     * to an enclosing BeanContext.
     */

    private class BCSSProxyProvider implements BeanContextServiceProvider, BeanContextServiceRevokedListener {
	
	BCSSProxyProvider(BeanContextServices bcs, Class sc) { 
	    super();
	
	    serviceClass = sc;
	    nestingCtxt  = bcs;
	}

	public Object getService(BeanContextServices bcs, Object requestor, Class serviceClass, Object serviceSelector) {
	    Object service = null;

	    try {
		service = nestingCtxt.getService(bcs, requestor, serviceClass, serviceSelector, this);
	    } catch (TooManyListenersException tmle) {
		return null;
	    }
	
	    refCnt++;

	    return service;
	}

	public void releaseService(BeanContextServices bcs, Object requestor, Object service) {
	    nestingCtxt.releaseService(BeanContextServicesSupport.this.getBeanContextServicesPeer(), requestor, service);

	    if (--refCnt == 0)
		revokeService(serviceClass, this, false);
	}

	public Iterator getCurrentServiceSelectors(BeanContextServices bcs, Class serviceClass) {
	    return nestingCtxt.getCurrentServiceSelectors(serviceClass);
	}

	public void serviceRevoked(BeanContextServiceRevokedEvent bcsre) {
	    BeanContextServicesSupport.this.getBeanContextServicesPeer().revokeService(serviceClass, this, bcsre.isCurrentServiceInvalidNow());
	}

	/*
	 * fields
	 */

	private Class		    serviceClass;
	private BeanContextServices nestingCtxt;
	private int   		    refCnt;
    }

    /************************************************************************/

    /**
     * obtain a service which may be delegated
     */

     public synchronized Object getService(BeanContextChild child, Object requestor, Class serviceClass, Object serviceSelector, BeanContextServiceRevokedListener bcsrl) throws TooManyListenersException {
	Object              service = null;
	BCSSChild           bcsc    = (BCSSChild)children.get(child);
	BeanContextServices bcssp   = getBeanContextServicesPeer();

	if (bcsc == null) return null; // not a child ...

	if (requestor == null) throw new IllegalArgumentException("null actual for requestor");

	if (bcsrl == null) throw new IllegalArgumentException("null actual for listener");
	BCSSServiceProvider bcsssp = (BCSSServiceProvider)services.get(serviceClass);

	boolean tryToDelegateService = false;

	if (bcsssp != null) {
	    BeanContextServiceProvider bcsp = bcsssp.getServiceProvider();
	    service = bcsp.getService(bcssp, requestor, serviceClass, serviceSelector);
	    if (service != null) {
		try {
	            bcsc.usingService(requestor, service, serviceClass, bcsp, bcsrl);
		} catch (TooManyListenersException tmle) {
		    bcsp.releaseService(bcssp, requestor, service);
		    throw tmle;
		}

		return service;
	    } else
		tryToDelegateService = true;
	} else 
	    tryToDelegateService = true;


	if (tryToDelegateService) {
	    BeanContextServices bcs = null;

	    // see if current parent provides services ...

	    try {
		bcs = (BeanContextServices)getBeanContext();
	    } catch (ClassCastException cce) {
		return null; // nope ...
	    }

	    if (bcs == null) return null;

	    synchronized(bcs) {

		BCSSProxyProvider proxy = new BCSSProxyProvider(bcs, serviceClass);

	        // try to delegate ...

	        service = bcs.getService(bcssp, requestor, serviceClass, serviceSelector, proxy);
	
	        if (service != null) {
		    try {
			bcsc.usingService(requestor, service, serviceClass, proxy, bcsrl);
		    } catch (TooManyListenersException tmle) {
		        bcs.releaseService(bcssp, requestor, service);
		        throw tmle;
		    }

		    addService(serviceClass, proxy, false);

		    return service;
	        }
	    }
	}

        return null;
     }

    /**
     * release a service
     */

    public synchronized void releaseService(BeanContextChild child, Object requestor, Object service) {
	BCSSChild bcsc = (BCSSChild)children.get(child);

	if (bcsc == null) return; // not a child ...

	bcsc.releaseService(requestor, service);
    }

    /**
     * @return an iterator for all the currently registered service classes.
     */

    public synchronized Iterator getCurrentServiceClasses() {
	return new BCSIterator(services.keySet().iterator());
    }

    /**
     * @return an iterator for all the currently available service selectors
     * (if any) available for the specified service.
     */

    public synchronized Iterator getCurrentServiceSelectors(Class serviceClass) {
	BeanContextServiceProvider bcsp = (BeanContextServiceProvider)services.get(serviceClass);
	
	return bcsp != null ? new BCSIterator(bcsp.getCurrentServiceSelectors(getBeanContextServicesPeer(), serviceClass)) : null;
    }

    /**
     * BeanContextServicesListener callback, propagates event to all 
     * currently registered listeners and BeanContextServices children,
     * if this BeanContextService does not already implement this service
     * itself.
     *
     * subclasses may override or envelope this method to implement their
     * own propagation semantics.
     */

     public synchronized void serviceAvailable(BeanContextServiceAvailableEvent bcssae) {
	if (services.containsKey(bcssae.getServiceClass())) return;

	fireServiceAdded(bcssae);

	Iterator i = children.keySet().iterator();

	while (i.hasNext()) {
	    Object c = i.next();

	    if (c instanceof BeanContextServices) {
		((BeanContextServicesListener)c).serviceAvailable(bcssae);
	    }
	}
     }

    /**
     * @return the BeanContextServicesListener (if any) of the specified child
     */

    protected static final BeanContextServicesListener getChildBeanContextServicesListener(Object child) {
	try {
	    return (BeanContextServicesListener)child;
	} catch (ClassCastException cce) {
	    return null;
	}
    }

    /**
     * called from superclass child removal operations after a child
     * has been successfully removed. called with child synchronized.
     *
     * This subclass uses this hook to immediately revoke any services
     * being used by this child if it is a BeanContextChild.
     *
     * subclasses may envelope this method in order to implement their
     * own child removal side-effects.
     */

    protected void childJustRemovedHook(Object child, BCSChild bcsc) {
	BCSSChild bcssc = (BCSSChild)bcsc;

	bcssc.cleanupReferences();
    }

    /**
     * called from setBeanContext to notify a BeanContextChild
     * to release resources obtained from the nesting BeanContext.
     * 
     * This method revokes any services obtained from its parent.
     * 
     * subclasses may envelope this method to implement their own semantics.
     */

    protected synchronized void releaseBeanContextResources() {
	if (children.isEmpty()) return;

	Iterator i = services.entries().iterator();

	Object[] bcssc = children.values().toArray();

	while (i.hasNext()) {
	    Map.Entry entry  = (Map.Entry)i.next();

	    if (entry.getValue() instanceof BCSSProxyProvider) {
		for (int j = 0; j < bcssc.length; j++) {
		    ((BCSSChild)bcssc[j]).revokeService((Class)entry.getKey(), true);
		}
	    }
	}
    }

    /**
     * Fire a BeanContextServiceEvent notifying of a new service
     */

    protected final void fireServiceAdded(Class serviceClass) {
	BeanContextServiceAvailableEvent bcssae = new BeanContextServiceAvailableEvent(getBeanContextServicesPeer(), serviceClass); 

	fireServiceAdded(bcssae);
    }

    protected final void fireServiceAdded(BeanContextServiceAvailableEvent bcssae) {
	Object[]		         copy;

	synchronized (bcsListeners) { copy = bcsListeners.toArray(); }

	for (int i = 0; i < copy.length; i++) {
	    ((BeanContextServicesListener)copy[i]).serviceAvailable(bcssae);
	}
    }

    /**
     * Fire a BeanContextServiceEvent notifying of a service being revoked
     */

    protected final void fireServiceRevoked(Class serviceClass, boolean revokeNow) {
	Object[]		       copy;
	BeanContextServiceRevokedEvent bcsre = new BeanContextServiceRevokedEvent(getBeanContextServicesPeer(), serviceClass, revokeNow); 

	synchronized (bcsListeners) { copy = bcsListeners.toArray(); }

	for (int i = 0; i < copy.length; i++) {
	    ((BeanContextServicesListener)copy[i]).serviceRevoked(bcsre);
	}
   }

    /**
     * called from BeanContextSupport writeObject before it serializes the
     * children ...
     *
     * This class will serialize any Serializable BeanContextServiceProviders
     * herein.
     *
     * subclasses may envelope this method to insert their own serialization
     * processing that has to occur prior to serialization of the children
     */

    protected synchronized void bcsPreSerializationHook(ObjectOutputStream oos) throws IOException {

	oos.writeInt(serializable);

	if (serializable <= 0) return;


	Iterator i = services.entries().iterator();

	int count = 0;

	while (i.hasNext() && count < serializable) {
	    Map.Entry 		entry = (Map.Entry)i.next();
	    BCSSServiceProvider bcsp  = null;

	     try {
		bcsp = (BCSSServiceProvider)entry.getValue();
	     } catch (ClassCastException cce) {
		continue;
	     }

	     if (bcsp.getServiceProvider() instanceof Serializable) {
		oos.writeObject(entry.getKey());
		oos.writeObject(bcsp);
		count++;
	     }
	}

	if (count != serializable) 
	    throw new IOException("wrote different number of service providers than expected");
    }

    /**
     * called from BeanContextSupport readObject before it deserializes the
     * children ...
     *
     * This class will deserialize any Serializable BeanContextServiceProviders
     * serialized earlier thus making them available to the children when they
     * deserialized.
     *
     * subclasses may envelope this method to insert their own serialization
     * processing that has to occur prior to serialization of the children
     */

    protected synchronized void bcsPreDeserializationHook(ObjectInputStream ois) throws IOException, ClassNotFoundException {

	serializable = ois.readInt();

	int count = serializable;

	while (count > 0) {
	    services.put(ois.readObject(), ois.readObject());
	    count--;
	} 
    }
 
    /**
     * serialize the instance
     */

    private synchronized void writeObject(ObjectOutputStream oos) throws IOException {
	oos.defaultWriteObject();

	serialize(oos, (Collection)bcsListeners);
    }

    /**
     * deserialize the instance
     */

    private synchronized void readObject(ObjectInputStream ois) throws IOException, ClassNotFoundException {

	ois.defaultReadObject();

	deserialize(ois, (Collection)bcsListeners);
    }

    
    /*
     * fields
     */

    protected transient HashMap  services;
    private   transient int 	 serializable = 0; // custom serialized

    protected transient ArrayList bcsListeners;
}
