/*
 * @(#)RMISocketFactory.java	1.10 98/03/18
 *
 * Copyright 1996, 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package java.rmi.server;

import java.io.*;
import java.net.*;

/**
 * An <code>RMISocketFactory</code> instance is used by the RMI runtime
 * in order to obtain client and server sockets for RMI calls.  An
 * application may use the <code>setSocketFactory</code> method to
 * request that the RMI runtime use its socket factory instance
 * instead of the default implementation.<p>
 *
 * The default socket factory implementation used goes through a
 * three-tiered approach to creating client sockets. First, a direct
 * socket connection to the remote VM is attempted.  If that fails
 * (due to a firewall), the runtime uses HTTP with the explicit port
 * number of the server.  If the firewall does not allow this type of
 * communication, then HTTP to a cgi-bin script on the server is used
 * to POST the RMI call.<p>
 *
 * @version 1.10, 03/18/98
 * @author  Ann Wollrath
 * @author  Peter Jones
 * @since   JDK1.1
 */
public abstract class RMISocketFactory {

    /** Client/server socket factory to be used by RMI runtime */
    private static RMISocketFactory factory = null;
    /** default socket factory used by this RMI implementation */
    private static RMISocketFactory defaultSocketFactory;
    /** Handler for socket creation failure */
    private static RMIFailureHandler handler = null;

    /**
     * Create a client socket connected to the specified host and port.
     */
    public abstract Socket createSocket(String host, int port)
	throws IOException;

    /**
     * Create a client socket connected to the specified host and port,
     * using the requested socket type.
     */
    public Socket createSocket(String host, int port, SocketType type)
	throws IOException
    {
	throw new IOException("protocol not supported");
    }

    /**
     * Create a server socket on the specified port (port 0 indicates
     * an anonymous port).
     */
    public abstract ServerSocket createServerSocket(int port)
	throws IOException;

    /**
     * Create a server socket on the specified port (port 0 indicates
     * an anonymous port), using the requested socket type.
     */
    public ServerSocket createServerSocket(int port, SocketType type)
	throws IOException
    {
	throw new IOException("protocol not supported");
    }

    /**
     * Set the socket factory from which RMI gets sockets. The RMI
     * socket factory can only be set once. Note: The RMISocketFactory
     * may only be set if the current security manager allows setting
     * a socket factory; if disallowed, a SecurityException will be
     * thrown.
     */
    public synchronized static void setSocketFactory(RMISocketFactory fac)
	throws IOException
    {
    	if (factory != null) {
	    throw new SocketException("factory already defined");
	}
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkSetFactory();
	}
	factory = fac;
    }

    /**
     * Returns the socket factory used by RMI.
     */
    public synchronized static RMISocketFactory getSocketFactory() 
    {
	return factory;
    }

    /**
     * Returns a reference to the default socket factory used
     * by this RMI implementation.  This will be the factory used
     * by the RMI runtime when <code>getSocketFactory</code>
     * returns null.
     */
    public synchronized static RMISocketFactory getDefaultSocketFactory() {
	if (defaultSocketFactory == null) {
	    defaultSocketFactory =
		new sun.rmi.transport.proxy.RMIMasterSocketFactory();
	}
	return defaultSocketFactory;
    }

    /**
     * Set the failure handler to be called by the RMI runtime if
     * socket creation fails.  The default implementation of this
     * handler returns false (thus recreation of sockets is not
     * attempted by the runtime).
     */
    public synchronized static void setFailureHandler(RMIFailureHandler fh) 
    {
	handler = fh;
    }

    /**
     * Returns the handler for socket creation failure.
     */
    public synchronized static RMIFailureHandler getFailureHandler()
    {
	return handler;
    }
}
