/*
 * @(#)RevokedCertificate.java	1.5 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
 
package java.security.cert;

import java.math.BigInteger;
import java.util.Date;
import java.util.Set;

/**
 * <p>Abstract class for a revoked certificate in a CRL (Certificate 
 * Revocation List).
 *
 * The ASN.1 definition for <em>revokedCertificates</em> is:
 * <pre>
 * revokedCertificates    SEQUENCE OF SEQUENCE  {
 *     userCertificate    CertificateSerialNumber,
 *     revocationDate     ChoiceOfTime,
 *     crlEntryExtensions Extensions OPTIONAL
 *                        -- if present, must be v2
 * }  OPTIONAL
 *<p>
 * CertificateSerialNumber  ::=  INTEGER
 *<p>
 * Extensions  ::=  SEQUENCE SIZE (1..MAX) OF Extension
 *<p>
 * Extension  ::=  SEQUENCE  {
 *     extnId        OBJECT IDENTIFIER,
 *     critical      BOOLEAN DEFAULT FALSE,
 *     extnValue     OCTET STRING
 *                   -- contains a DER encoding of a value
 *                   -- of the type registered for use with
 *                   -- the extnId object identifier value
 * }
 * </pre>
 * 
 * @see X509CRL
 *
 * @author Hemma Prafullchandra
 * @version 1.5 98/03/18
 */

public abstract class RevokedCertificate implements X509Extension {

    /**
     * Gets the serial number for this RevokedCertificate,
     * the <em>userCertificate</em>.
     * 
     * @return the serial number.
     */
    public abstract BigInteger getSerialNumber();

    /**
     * Gets the revocation date for this RevokedCertificate,
     * the <em>revocationDate</em>.
     * 
     * @return the revocation date.
     */
    public abstract Date getRevocationDate();

    /**
     * Returns true if this revoked certificate entry has
     * extensions.
     *
     * @return true if this entry has extensions, false otherwise.
     */
    public abstract boolean hasExtensions();

    /**
     * Returns a string representation of this revoked certificate.
     *
     * @return a string representation of this revoked certificate.
     */
    public abstract String toString();
}
